import numpy as np
import traceback
import logging
import datetime
import webvtt
import copy
import attr
import json
import os
import warnings

from PIL import Image
from enum import Enum

from .. import entities, PlatformException, repositories, ApiClient, exceptions

logger = logging.getLogger(name='dtlpy')


class ExportVersion(str, Enum):
    V1 = "V1"
    V2 = "V2"


class AnnotationStatus(str, Enum):
    ISSUE = "issue"
    APPROVED = "approved"
    REVIEW = "review"
    CLEAR = "clear"


class AnnotationType(str, Enum):
    BOX = "box"
    CUBE = "cube"
    CUBE3D = "cube_3d"
    CLASSIFICATION = "class"
    COMPARISON = "comparison"
    ELLIPSE = "ellipse"
    NOTE = "note"
    POINT = "point"
    POLYGON = "segment"
    POLYLINE = "polyline"
    POSE = "pose"
    SEGMENTATION = "binary"
    SUBTITLE = "subtitle"
    TEXT = "text_mark"
    GIS = "gis"
    SEMANTIC_3D = "ref_semantic_3d"
    POLYLINE_3D = "polyline_3d"
    FREE_TEXT = "text"
    REF_IMAGE = "ref_image"


class ViewAnnotationOptions(str, Enum):
    """ The Annotations file types to download (JSON, MASK, INSTANCE, ANNOTATION_ON_IMAGE, VTT, OBJECT_ID).

    .. list-table::
       :widths: 15 150
       :header-rows: 1

       * - State
         - Description
       * - JSON
         - Dataloop json format
       * - MASK
         - PNG file that contains drawing annotations on it
       * - INSTANCE
         - An image file that contains 2D annotations
       * - ANNOTATION_ON_IMAGE
         - The source image with the annotations drawing in it
       * - VTT
         - An text file contains supplementary information about a web video
       * - OBJECT_ID
         - An image file that contains 2D annotations

    """
    JSON = "json"
    MASK = "mask"
    INSTANCE = "instance"
    ANNOTATION_ON_IMAGE = "img_mask"
    VTT = "vtt"
    OBJECT_ID = "object_id"


@attr.s
class Annotation(entities.BaseEntity):
    """
    Annotations object
    """
    # platform
    id = attr.ib()
    url = attr.ib(repr=False)
    item_url = attr.ib(repr=False)
    _item = attr.ib(repr=False)
    item_id = attr.ib()
    creator = attr.ib()
    created_at = attr.ib()
    updated_by = attr.ib(repr=False)
    updated_at = attr.ib(repr=False)
    type = attr.ib()
    source = attr.ib(repr=False)
    dataset_url = attr.ib(repr=False)

    # api
    _platform_dict = attr.ib(repr=False)
    # meta
    metadata = attr.ib(repr=False)
    fps = attr.ib(repr=False)
    hash = attr.ib(default=None, repr=False)
    dataset_id = attr.ib(default=None, repr=False)
    status = attr.ib(default=None, repr=False)
    object_id = attr.ib(default=None, repr=False)
    automated = attr.ib(default=None, repr=False)
    item_height = attr.ib(default=None)
    item_width = attr.ib(default=None)
    label_suggestions = attr.ib(default=None)

    # annotation definition
    _annotation_definition = attr.ib(default=None, repr=False, type=entities.BaseAnnotationDefinition)

    # snapshots
    frames = attr.ib(default=None, repr=False)
    current_frame = attr.ib(default=0, repr=False)

    # video attributes
    _end_frame = attr.ib(default=0, repr=False)
    _end_time = attr.ib(default=0, repr=False)
    _start_frame = attr.ib(default=0)
    _start_time = attr.ib(default=0)

    # sdk
    _dataset = attr.ib(repr=False, default=None)
    _datasets = attr.ib(repr=False, default=None)
    _annotations = attr.ib(repr=False, default=None)
    __client_api = attr.ib(default=None, repr=False)
    _items = attr.ib(repr=False, default=None)
    _recipe_1_attributes = attr.ib(repr=False, default=None)

    ############
    # Platform #
    ############

    @property
    def annotation_definition(self):
        return self._annotation_definition

    @annotation_definition.setter
    def annotation_definition(self, ann_def):
        if ann_def is not None:
            ann_def._annotation = self
        self._annotation_definition = ann_def

    @property
    def createdAt(self):
        return self.created_at

    @property
    def updatedAt(self):
        return self.updated_at

    @property
    def updatedBy(self):
        return self.updated_by

    @property
    def _client_api(self) -> ApiClient:
        if self.__client_api is None:
            if self._item is None:
                raise PlatformException('400',
                                        'This action cannot be performed without an item entity. Please set item')
            else:
                self.__client_api = self._item._client_api
        assert isinstance(self.__client_api, ApiClient)
        return self.__client_api

    @property
    def dataset(self):
        if self._dataset is None:
            if self._item is not None:
                # get from item
                self._dataset = self._item.dataset
            else:
                # get directly
                self._dataset = self.datasets.get(dataset_id=self.dataset_id)
        assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    @property
    def item(self):
        if self._item is None:
            self._item = self.items.get(item_id=self.item_id)
        assert isinstance(self._item, entities.Item)
        return self._item

    @property
    def annotations(self):
        if self._annotations is None:
            self._annotations = repositories.Annotations(client_api=self._client_api, item=self._item)
        assert isinstance(self._annotations, repositories.Annotations)
        return self._annotations

    @property
    def datasets(self):
        if self._datasets is None:
            self._datasets = repositories.Datasets(client_api=self._client_api)
        assert isinstance(self._datasets, repositories.Datasets)
        return self._datasets

    @property
    def items(self):
        if self._items is None:
            if self._datasets is not None:
                self._items = self._dataset.items
            elif self._item is not None:
                self._items = self._item.items
            else:
                self._items = repositories.Items(client_api=self._client_api, dataset=self._dataset)
        assert isinstance(self._items, repositories.Items)
        return self._items

    #########################
    # Annotation Properties #
    #########################
    @property
    def parent_id(self):
        try:
            parent_id = self.metadata['system']['parentId']
        except KeyError:
            parent_id = None
        return parent_id

    @parent_id.setter
    def parent_id(self, parent_id):
        if 'system' not in self.metadata:
            self.metadata['system'] = dict()
        self.metadata['system']['parentId'] = parent_id

    @property
    def coordinates(self):
        color = None
        if self.type in ['binary']:
            color = self.color
        coordinates = self.annotation_definition.to_coordinates(color=color)
        return coordinates

    @property
    def start_frame(self):
        return self._start_frame

    @start_frame.setter
    def start_frame(self, val):
        if not isinstance(val, float) and not isinstance(val, int):
            raise ValueError('Must input a valid number')
        self._start_frame = round(val)
        self._start_time = val / self.fps if self.fps else 0
        self.frames.start = self._start_frame

    @property
    def end_frame(self):
        return self._end_frame

    @end_frame.setter
    def end_frame(self, val):
        if not isinstance(val, float) and not isinstance(val, int):
            raise ValueError('Must input a valid number')
        self._end_frame = round(val)
        self._end_time = val / self.fps if self.fps else 0
        self.frames.end = self._end_frame

    @property
    def start_time(self):
        return self._start_time

    @start_time.setter
    def start_time(self, val):
        if not isinstance(val, float) and not isinstance(val, int):
            raise ValueError('Must input a valid number')
        self._start_frame = round(val * self.fps if self.fps else 0)
        self._start_time = val
        self.frames.start = self._start_frame

    @property
    def end_time(self):
        return self._end_time

    @end_time.setter
    def end_time(self, val):
        if not isinstance(val, float) and not isinstance(val, int):
            raise ValueError('Must input a valid number')
        self._end_frame = round(val * self.fps if self.fps else 0)
        self._end_time = val
        self.frames.end = self._end_frame

    @property
    def x(self):
        return self.annotation_definition.x

    @property
    def y(self):
        return self.annotation_definition.y

    @property
    def rx(self):
        if self.annotation_definition.type == 'ellipse':
            return self.annotation_definition.rx
        else:
            return None

    @property
    def ry(self):
        if self.annotation_definition.type == 'ellipse':
            return self.annotation_definition.ry
        else:
            return None

    @property
    def angle(self):
        if self.annotation_definition.type in ['ellipse', 'cube', 'box']:
            return self.annotation_definition.angle
        else:
            return None

    @property
    def messages(self):
        if hasattr(self.annotation_definition, 'messages'):
            return self.annotation_definition.messages
        else:
            return None

    @messages.setter
    def messages(self, messages):
        if self.type == 'note':
            self.annotation_definition.messages = messages
        else:
            raise PlatformException('400', 'Annotation of type {} does not have attribute messages'.format(self.type))

    def add_message(self, body: str = None):
        if self.type == 'note':
            return self.annotation_definition.add_message(body=body)
        else:
            raise PlatformException('400', 'Annotation of type {} does not have method add_message'.format(self.type))

    @property
    def geo(self):
        return self.annotation_definition.geo

    @geo.setter
    def geo(self, geo):
        self.annotation_definition.geo = geo

    @property
    def top(self):
        return self.annotation_definition.top

    @top.setter
    def top(self, top):
        self.annotation_definition.top = top

    @property
    def bottom(self):
        return self.annotation_definition.bottom

    @bottom.setter
    def bottom(self, bottom):
        self.annotation_definition.bottom = bottom

    @property
    def left(self):
        return self.annotation_definition.left

    @left.setter
    def left(self, left):
        self.annotation_definition.left = left

    @property
    def right(self):
        return self.annotation_definition.right

    @right.setter
    def right(self, right):
        self.annotation_definition.right = right

    @property
    def height(self):
        return self.annotation_definition.height

    @height.setter
    def height(self, height):
        self.annotation_definition.height = height

    @property
    def width(self):
        return self.annotation_definition.width

    @width.setter
    def width(self, width):
        self.annotation_definition.width = width

    @property
    def description(self):
        return self.annotation_definition.description

    @description.setter
    def description(self, description):
        if description is None:
            description = ""
        if not isinstance(description, str):
            raise ValueError("Description must get string")
        self.annotation_definition.description = description

    @property
    def last_frame(self):
        if len(self.frames.actual_keys()) == 0:
            return 0
        return max(self.frames.actual_keys())

    @property
    def label(self):
        return self.annotation_definition.label

    @label.setter
    def label(self, label):
        self.annotation_definition.label = label

    @property
    def attributes(self):
        return self.annotation_definition.attributes

    @attributes.setter
    def attributes(self, attributes):
        self.annotation_definition.attributes = attributes

    @property
    def color(self):
        # if "dataset" is not in self - this will always get the dataset
        try:
            colors = self.dataset._get_ontology().color_map
        except (exceptions.BadRequest, exceptions.NotFound):
            colors = None
            logger.warning('Cant get dataset for annotation color. using default.')
        if colors is not None and self.label in colors:
            color = colors[self.label]
        else:
            if self.type == 'binary' and self.annotation_definition._color is not None:
                color = self.annotation_definition._color
            else:
                color = (255, 255, 255)
        return color

    @color.setter
    def color(self, color):
        if self.type == 'binary':
            if not isinstance(color, tuple) or len(color) != 3:
                raise ValueError("Color must get tuple of length 3")
            self.annotation_definition._color = color
        else:
            raise exceptions.BadRequest(
                status_code='400',
                message='Invalid annotation type - Updating color is only available to binary annotation'
            )

    ####################
    # frame attributes #
    ####################
    @property
    def frame_num(self):
        if len(self.frames.actual_keys()) > 0:
            return self.current_frame
        else:
            return self.start_frame

    @frame_num.setter
    def frame_num(self, frame_num):
        if frame_num != self.current_frame:
            self.frames[self.current_frame].frame_num = frame_num
            self.frames[frame_num] = self.frames[self.current_frame]
            self.frames.pop(self.current_frame)

    @property
    def fixed(self):
        if len(self.frames.actual_keys()) > 0:
            return self.frames[self.current_frame].fixed
        else:
            return False

    @fixed.setter
    def fixed(self, fixed):
        if len(self.frames.actual_keys()) > 0:
            self.frames[self.current_frame].fixed = fixed

    @property
    def object_visible(self):
        if len(self.frames.actual_keys()) > 0:
            return self.frames[self.current_frame].object_visible
        else:
            return False

    @object_visible.setter
    def object_visible(self, object_visible):
        if len(self.frames.actual_keys()) > 0:
            self.frames[self.current_frame].object_visible = object_visible

    @property
    def is_video(self):
        if len(self.frames.actual_keys()) == 0:
            return False
        else:
            return True

    ##################
    # entity methods #
    ##################
    def update_status(self, status: AnnotationStatus = AnnotationStatus.ISSUE):
        """
        Set status on annotation

        **Prerequisites**: You must have an item that has been annotated. You must have the role of an *owner* or *developer* or be assigned a task that includes that item as an *annotation manager*.

        :param str status: can be AnnotationStatus.ISSUE, AnnotationStatus.APPROVED, AnnotationStatus.REVIEW, AnnotationStatus.CLEAR
        :return: Annotation object
        :rtype: dtlpy.entities.annotation.Annotation

        **Example**:

        .. code-block:: python

            annotation = annotation.update_status(status=dl.AnnotationStatus.ISSUE)
        """
        return self.annotations.update_status(annotation=self, status=status)

    def delete(self):
        """
        Remove an annotation from item

        **Prerequisites**: You must have an item that has been annotated. You must have the role of an *owner* or *developer* or be assigned a task that includes that item as an *annotation manager* or *annotator*.

        :return: True if success
        :rtype: bool

        **Example**:

        .. code-block:: python

            is_deleted = annotation.delete()
        """
        if self.id is None:
            raise PlatformException(
                '400',
                'Cannot delete annotation because it was not fetched from platform and therefore does not have an id'
            )
        return self.annotations.delete(annotation_id=self.id)

    def update(self, system_metadata=False):
        """
        Update an existing annotation in host.

        **Prerequisites**: You must have an item that has been annotated. You must have the role of an *owner* or *developer* or be assigned a task that includes that item as an *annotation manager* or *annotator*.

        :param system_metadata: True, if you want to change metadata system
        :return: Annotation object
        :rtype: dtlpy.entities.annotation.Annotation

        **Example**:

        .. code-block:: python

            annotation = annotation.update()
        """
        return self.annotations.update(annotations=self,
                                       system_metadata=system_metadata)[0]

    def upload(self):
        """
        Create a new annotation in host

        **Prerequisites**: Any user can upload annotations.

        :return: Annotation entity
        :rtype: dtlpy.entities.annotation.Annotation
        """
        return self.annotations.upload(annotations=self)[0]

    def download(self,
                 filepath: str,
                 annotation_format: ViewAnnotationOptions = ViewAnnotationOptions.JSON,
                 height: float = None,
                 width: float = None,
                 thickness: int = 1,
                 with_text: bool = False,
                 alpha: float = 1):
        """
        Save annotation to file

        **Prerequisites**: Any user can upload annotations.

        :param str filepath: local path to where annotation will be downloaded to
        :param list annotation_format: options: list(dl.ViewAnnotationOptions)
        :param float height: image height
        :param float width: image width
        :param int thickness: line thickness
        :param bool with_text: get mask with text
        :param float alpha: opacity value [0 1], default 1
        :return: filepath
        :rtype: str

        **Example**:

        .. code-block:: python

            filepath = annotation.download(filepath='filepath', annotation_format=dl.ViewAnnotationOptions.MASK)
        """
        _, ext = os.path.splitext(filepath)
        if ext == '':
            if annotation_format == ViewAnnotationOptions.JSON:
                ext = '.json'
            else:
                if self.is_video:
                    ext = '.mp4'
                else:
                    ext = '.png'
            filepath = os.path.join(filepath, self.id + ext)
        os.makedirs(os.path.dirname(filepath), exist_ok=True)
        if annotation_format == ViewAnnotationOptions.JSON:
            with open(filepath, 'w') as f:
                json.dump(self.to_json(), f, indent=2)
        elif annotation_format == entities.ViewAnnotationOptions.VTT:
            _, ext = os.path.splitext(filepath)
            if ext != '.vtt':
                raise PlatformException(error='400', message='file extension must be vtt')
            vtt = webvtt.WebVTT()
            if self.type in ['subtitle']:
                s = str(datetime.timedelta(seconds=self.start_time))
                if len(s.split('.')) == 1:
                    s += '.000'
                e = str(datetime.timedelta(seconds=self.end_time))
                if len(e.split('.')) == 1:
                    e += '.000'
                caption = webvtt.Caption(
                    '{}'.format(s),
                    '{}'.format(e),
                    '{}'.format(self.coordinates['text'])
                )
                vtt.captions.append(caption)
            vtt.save(filepath)
        else:
            if self.is_video:
                entities.AnnotationCollection(item=self.item, annotations=[self])._video_maker(
                    input_filepath=None,
                    output_filepath=filepath,
                    annotation_format=annotation_format,
                    thickness=thickness,
                    alpha=alpha,
                    with_text=with_text
                )
            else:
                mask = self.show(thickness=thickness,
                                 alpha=alpha,
                                 with_text=with_text,
                                 height=height,
                                 width=width,
                                 annotation_format=annotation_format)
                img = Image.fromarray(mask.astype(np.uint8))
                img.save(filepath)
        return filepath

    def set_frame(self, frame):
        """
        Set annotation to frame state

        **Prerequisites**: Any user can upload annotations.

        :param int frame: frame number
        :return: True if success
        :rtype: bool

        **Example**:

        .. code-block:: python

            success = annotation.set_frame(frame=10)
        """
        if frame in self.frames:
            self.current_frame = frame
            self.annotation_definition = self.frames[frame].annotation_definition
            return True
        else:
            return False

    ############
    # Plotting #
    ############
    def show(self,
             image=None,
             thickness=None,
             with_text=False,
             height=None,
             width=None,
             annotation_format: ViewAnnotationOptions = ViewAnnotationOptions.MASK,
             color=None,
             label_instance_dict=None,
             alpha=1,
             frame_num=None
             ):
        """
        Show annotations
        mark the annotation of the image array and return it

        **Prerequisites**: Any user can upload annotations.

        :param image: empty or image to draw on
        :param int thickness: line thickness
        :param bool with_text: add label to annotation
        :param float height: height
        :param float width: width
        :param dl.ViewAnnotationOptions annotation_format: list(dl.ViewAnnotationOptions)
        :param tuple color: optional - color tuple
        :param label_instance_dict: the instance labels
        :param float alpha: opacity value [0 1], default 1
        :param int frame_num: for video annotation, show specific fame
        :return: list or single ndarray of the annotations

        **Exampls**:

        .. code-block:: python

            image = annotation.show(image='ndarray',
                            thickness=1,
                            annotation_format=dl.VIEW_ANNOTATION_OPTIONS_MASK,
                            )
        """
        try:
            import cv2
        except (ImportError, ModuleNotFoundError):
            logger.error(
                'Import Error! Cant import cv2. Annotations operations will be limited. import manually and fix errors')
            raise
        # sanity check for alpha
        if alpha > 1 or alpha < 0:
            raise ValueError(
                'input variable alpha in annotation.show() should be between 0 and 1. got: {!r}'.format(alpha))
        if height is None:
            if self._item is None or self._item.height is None:
                if image is None:
                    raise PlatformException(error='400', message='must provide item width and height')
                else:
                    height = image.shape[0]
            else:
                height = self._item.height
        if width is None:
            if self._item is None or self._item.width is None:
                if image is None:
                    raise PlatformException(error='400', message='must provide item width and height')
                else:
                    width = image.shape[1]
            else:
                width = self._item.width

        # try getting instance map from dataset, otherwise set to default
        if annotation_format in [entities.ViewAnnotationOptions.INSTANCE, entities.ViewAnnotationOptions.OBJECT_ID] and \
                label_instance_dict is None:
            if self._dataset is not None:
                label_instance_dict = self._dataset.instance_map
            else:
                if self._item is not None and self._item._dataset is not None:
                    label_instance_dict = self._item._dataset.instance_map
            if label_instance_dict is None:
                logger.warning("Couldn't set label_instance_dict in annotation.show(). All labels will be mapped to 1")
                label_instance_dict = dict()

        in_video_range = True
        if frame_num is not None:
            in_video_range = self.set_frame(frame_num)

        if self.is_video and not in_video_range:
            if image is None:
                image = self._get_default_mask(annotation_format=annotation_format,
                                               width=width,
                                               height=height,
                                               label_instance_dict=label_instance_dict)
            return image
        else:
            image = self._show_single_frame(image=image,
                                            thickness=thickness,
                                            alpha=alpha,
                                            with_text=with_text,
                                            height=height,
                                            width=width,
                                            annotation_format=annotation_format,
                                            color=color,
                                            label_instance_dict=label_instance_dict)
        return image

    def _show_single_frame(self,
                           image=None,
                           thickness=None,
                           with_text=False,
                           height=None,
                           width=None,
                           annotation_format: ViewAnnotationOptions = ViewAnnotationOptions.MASK,
                           color=None,
                           label_instance_dict=None,
                           alpha=None):
        """
        Show annotations
        mark the annotation of the single frame array and return it
        :param image: empty or image to draw on
        :param thickness: line thickness
        :param with_text: add label to annotation
        :param height: height
        :param width: width
        :param annotation_format: list(dl.ViewAnnotationOptions)
        :param color: optional - color tuple
        :param label_instance_dict: the instance labels
        :param alpha: opacity value [0 1], default 1
        :return: ndarray of the annotations
        """
        try:
            import cv2
        except (ImportError, ModuleNotFoundError):
            logger.error(
                'Import Error! Cant import cv2. Annotations operations will be limited. import manually and fix errors')
            raise
        # height/width
        if self.annotation_definition.type == 'cube_3d':
            logger.warning('the show for 3d_cube is not supported.')
            return image
        if annotation_format == entities.ViewAnnotationOptions.MASK:
            # create an empty mask
            if image is None:
                image = self._get_default_mask(annotation_format=annotation_format,
                                               width=width,
                                               height=height,
                                               label_instance_dict=label_instance_dict)
            else:
                if len(image.shape) == 2:
                    # image is gray
                    image = cv2.cvtColor(image, cv2.COLOR_GRAY2RGBA)
                elif image.shape[2] == 3:
                    image = cv2.cvtColor(image, cv2.COLOR_RGB2RGBA)
                elif image.shape[2] == 4:
                    pass
                else:
                    raise PlatformException(
                        error='1001',
                        message='Unknown image shape. expected depth: gray or RGB or RGBA. got: {}'.format(image.shape))
        elif annotation_format == entities.ViewAnnotationOptions.ANNOTATION_ON_IMAGE:
            if image is None:
                raise PlatformException(error='1001',
                                        message='Must input image with ANNOTATION_ON_IMAGE view option.')
        elif annotation_format in [entities.ViewAnnotationOptions.INSTANCE, entities.ViewAnnotationOptions.OBJECT_ID]:
            if image is None:
                # create an empty mask
                image = self._get_default_mask(annotation_format=annotation_format,
                                               width=width,
                                               height=height,
                                               label_instance_dict=label_instance_dict)
            else:
                if len(image.shape) != 2:
                    raise PlatformException(
                        error='1001',
                        message='Image shape must be 2d array when trying to draw instance on image')
            # create a dictionary of labels and ids
        else:
            raise PlatformException(error='1001',
                                    message='unknown annotations format: "{}". known formats: "{}"'.format(
                                        annotation_format, '", "'.join(list(entities.ViewAnnotationOptions))))
        # color
        if color is None:
            if annotation_format in [entities.ViewAnnotationOptions.MASK,
                                     entities.ViewAnnotationOptions.INSTANCE,
                                     entities.ViewAnnotationOptions.OBJECT_ID]:
                color = self._get_color_for_show(annotation_format=annotation_format,
                                                 alpha=alpha,
                                                 label_instance_dict=label_instance_dict)
            else:
                raise PlatformException(error='1001',
                                        message='unknown annotations format: {}. known formats: "{}"'.format(
                                            annotation_format, '", "'.join(list(entities.ViewAnnotationOptions))))

        return self.annotation_definition.show(image=image,
                                               thickness=thickness,
                                               with_text=with_text,
                                               height=height,
                                               width=width,
                                               annotation_format=annotation_format,
                                               color=color,
                                               alpha=alpha)

    def _get_color_for_show(self, annotation_format, alpha=1, label_instance_dict=None):
        if annotation_format == entities.ViewAnnotationOptions.MASK:
            color = self.color
            if len(color) == 3:
                color = color + (int(alpha * 255),)
        elif annotation_format == entities.ViewAnnotationOptions.INSTANCE:
            color = label_instance_dict.get(self.label, 1)
        elif annotation_format == entities.ViewAnnotationOptions.OBJECT_ID:
            if self.object_id is None:
                raise ValueError('Try to show object_id but annotation has no value. annotation id: {}'.format(self.id))
            color = int(self.object_id)
        else:
            raise ValueError('cant find color for the annotation format: {}'.format(annotation_format))
        return color

    def _get_default_mask(self, annotation_format, height, width, label_instance_dict):
        """
        Create a default mask take from the colors or instance map. tag should '$default'.
        Default value is zeros

        :param annotation_format: dl.AnnotationOptions to show
        :param height: item height
        :param width: item width
        :return:
        """
        if annotation_format == entities.ViewAnnotationOptions.MASK:
            try:
                colors = self.dataset._get_ontology().color_map
            except exceptions.BadRequest:
                colors = None
                logger.warning('Cant get dataset for annotation color. using default.')
            if colors is not None and '$default' in colors:
                default_color = colors['$default']
            else:
                default_color = None
            if default_color is None:
                default_color = (0, 0, 0, 0)
            if len(default_color) == 3:
                default_color += (255,)
            if len(default_color) != 4:
                raise ValueError('default color for show() mask should be with len 3 or 4 (RGB/RGBA)')
            image = np.full(shape=(height, width, 4), fill_value=default_color, dtype=np.uint8)
        elif annotation_format in [entities.ViewAnnotationOptions.INSTANCE,
                                   entities.ViewAnnotationOptions.OBJECT_ID]:
            default_color = 0
            if '$default' in label_instance_dict:
                default_color = label_instance_dict['$default']
            image = np.full(shape=(height, width), fill_value=default_color, dtype=np.uint8)
        else:
            raise ValueError('cant create default mask for annotation_format: {}'.format(annotation_format))
        return image

    #######
    # I/O #
    #######
    @classmethod
    def new(cls,
            item=None,
            annotation_definition=None,
            object_id=None,
            automated=True,
            metadata=None,
            frame_num=None,
            parent_id=None,
            start_time=None,
            item_height=None,
            item_width=None,
            end_time=None):
        """
        Create a new annotation object annotations

        **Prerequisites**: Any user can upload annotations.

        :param dtlpy.entities.item.Items item: item to annotate
        :param annotation_definition: annotation type object
        :param str object_id: object_id
        :param bool automated: is automated
        :param dict metadata: metadata
        :param int frame_num: optional - first frame number if video annotation
        :param str parent_id: add parent annotation ID
        :param start_time: optional - start time if video annotation
        :param float item_height: annotation item's height
        :param float item_width: annotation item's width
        :param end_time: optional - end time if video annotation
        :return: annotation object
        :rtype: dtlpy.entities.annotation.Annotation

        **Example**:

        .. code-block:: python

            annotation = annotation.new(item='item_entity,
                            annotation_definition=dl.Box(top=10,left=10,bottom=100, right=100,label='labelName'))
                            )
        """
        if frame_num is None:
            frame_num = 0

        if object_id is not None:
            if isinstance(object_id, int):
                object_id = '{}'.format(object_id)
            elif not isinstance(object_id, str) or not object_id.isnumeric():
                raise ValueError('Object id must be an int or a string containing only numbers.')

        # init annotations
        if metadata is None:
            metadata = dict()

        # add parent
        if parent_id is not None:
            if 'system' not in metadata:
                metadata['system'] = dict()
            metadata['system']['parentId'] = parent_id

        # add note status to metadata
        if annotation_definition is not None and annotation_definition.type == 'note':
            if 'system' not in metadata:
                metadata['system'] = dict()
            metadata['system']['status'] = annotation_definition.status

        # handle fps
        fps = None
        if item is not None:
            if item.fps is not None:
                fps = item.fps
            elif item.mimetype is not None and 'audio' in item.mimetype:
                fps = 1000

        # get type
        ann_type = None
        if annotation_definition is not None:
            ann_type = annotation_definition.type

        # dataset
        dataset_url = None
        dataset_id = None
        if item is not None:
            dataset_url = item.dataset_url
            dataset_id = item.dataset_id

        if start_time is None:
            if fps is not None and frame_num is not None:
                start_time = frame_num / fps if fps != 0 else 0
            else:
                start_time = 0
        if end_time is None:
            end_time = start_time

        if frame_num is None:
            frame_num = 0

        # frames
        frames = entities.ReflectDict(
            value_type=FrameAnnotation,
            on_access=Annotation.on_access,
            start=frame_num,
            end=frame_num
        )

        res = cls(
            # platform
            id=None,
            url=None,
            item_url=None,
            item=item,
            item_id=None,
            creator=None,
            created_at=None,
            updated_by=None,
            updated_at=None,
            object_id=object_id,
            type=ann_type,
            dataset_url=dataset_url,
            dataset_id=dataset_id,
            item_height=item_height,
            item_width=item_width,

            # meta
            metadata=metadata,
            fps=fps,
            status=None,
            automated=automated,

            # snapshots
            frames=frames,

            # video only attributes
            end_frame=frame_num,
            end_time=end_time,
            start_frame=frame_num,
            start_time=start_time,

            # temp
            platform_dict=dict(),
            source='sdk'
        )

        if annotation_definition:
            res.annotation_definition = annotation_definition

        if annotation_definition and annotation_definition.attributes is not None:
            res.attributes = annotation_definition.attributes

        return res

    def add_frames(self,
                   annotation_definition,
                   frame_num=None,
                   end_frame_num=None,
                   start_time=None,
                   end_time=None,
                   fixed=True,
                   object_visible=True):
        """
        Add a frames state to annotation

        **Prerequisites**: Any user can upload annotations.

        :param annotation_definition: annotation type object - must be same type as annotation
        :param int frame_num: first frame number
        :param int end_frame_num: last frame number
        :param start_time: starting time for video
        :param end_time: ending time for video
        :param bool fixed: is fixed
        :param bool object_visible: does the annotated object is visible
        :return:

        **Example**:

        .. code-block:: python

            annotation.add_frames(frame_num=10,
                            annotation_definition=dl.Box(top=10,left=10,bottom=100, right=100,label='labelName'))
                            )
        """
        # handle fps
        if self.fps is None:
            if self._item is not None:
                if self._item.fps is not None:
                    self.fps = self._item.fps
        if self.fps is None:
            raise PlatformException('400', 'Annotation must have fps in order to perform this action')

        if frame_num is None:
            frame_num = int(np.round(start_time * self.fps))

        if end_frame_num is None:
            if end_time is not None:
                end_frame_num = int(np.round(end_time * self.fps))
            else:
                end_frame_num = frame_num

        for frame in range(frame_num, end_frame_num + 1):
            self.add_frame(annotation_definition=annotation_definition,
                           frame_num=frame,
                           fixed=fixed,
                           object_visible=object_visible)

    def add_frame(self,
                  annotation_definition,
                  frame_num=None,
                  fixed=True,
                  object_visible=True):
        """
        Add a frame state to annotation

        :param annotation_definition: annotation type object - must be same type as annotation
        :param int frame_num: frame number
        :param bool fixed: is fixed
        :param bool object_visible: does the annotated object is visible
        :return: True if success
        :rtype: bool

        **Example**:

        .. code-block:: python

            success = annotation.add_frame(frame_num=10,
                            annotation_definition=dl.Box(top=10,left=10,bottom=100, right=100,label='labelName'))
                            )
        """
        # handle fps
        if self.fps is None:
            if self._item is not None:
                if self._item.fps is not None:
                    self.fps = self._item.fps
        if self.fps is None:
            raise PlatformException('400', 'Annotation must have fps in order to perform this action')

        # if this is first frame
        if self.annotation_definition is None:

            if frame_num is None:
                frame_num = 0

            frame = FrameAnnotation.new(annotation_definition=annotation_definition,
                                        frame_num=frame_num,
                                        fixed=fixed,
                                        object_visible=object_visible,
                                        annotation=self)

            self.frames[frame_num] = frame
            self.set_frame(frame_num)
            self.current_frame = frame_num
            self.end_frame = frame_num
            self.type = annotation_definition.type

            return True

        # check if type matches annotation
        if not isinstance(annotation_definition, type(self.annotation_definition)):
            raise PlatformException('400', 'All frames in annotation must have same type')

        # find frame number
        if frame_num is None:
            frame_num = self.last_frame + 1
        elif frame_num < self.start_frame:
            self.start_frame = frame_num

        # add frame to annotation
        if not self.is_video:
            # create first frame from annotation definition
            frame = FrameAnnotation.new(annotation_definition=self.annotation_definition,
                                        frame_num=self.last_frame,
                                        fixed=fixed,
                                        object_visible=object_visible,
                                        annotation=self)

            self.frames[self.start_frame] = frame

        # create new time annotations
        frame = FrameAnnotation.new(annotation_definition=annotation_definition,
                                    frame_num=frame_num,
                                    fixed=fixed,
                                    object_visible=object_visible,
                                    annotation=self)

        self.frames[frame_num] = frame
        if not self.frames[frame_num].fixed and self.start_frame < frame_num:
            frame_last = self.frames[frame_num - 1].to_snapshot()
            frame_current = self.frames[frame_num].to_snapshot()
            frame_last.pop('frame')
            frame_current.pop('frame')
            if frame_last == frame_current:
                self.frames[frame_num]._interpolation = True
        self.end_frame = max(self.last_frame, frame_num)

        return True

    @staticmethod
    def _protected_from_json(_json,
                             item=None,
                             client_api=None,
                             annotations=None,
                             is_video=None,
                             fps=None,
                             item_metadata=None,
                             dataset=None):
        """
        Same as from_json but with try-except to catch if error

        :param _json: platform json
        :param item: item
        :param client_api: ApiClient entity
        :param annotations:
        :param is_video:
        :param fps:
        :param item_metadata:
        :param dataset
        :return: annotation object
        """
        try:
            annotation = Annotation.from_json(_json=_json,
                                              item=item,
                                              client_api=client_api,
                                              annotations=annotations,
                                              is_video=is_video,
                                              fps=fps,
                                              item_metadata=item_metadata,
                                              dataset=dataset)
            status = True
        except Exception:
            annotation = traceback.format_exc()
            status = False
        return status, annotation

    @classmethod
    def from_json(cls,
                  _json,
                  item=None,
                  client_api=None,
                  annotations=None,
                  is_video=None,
                  fps=None,
                  item_metadata=None,
                  dataset=None,
                  is_audio=None):
        """
        Create an annotation object from platform json

        :param dict _json: platform json
        :param dtlpy.entities.item.Item item: item
        :param client_api: ApiClient entity
        :param annotations:
        :param bool is_video: is video
        :param fps: video fps
        :param item_metadata: item metadata
        :param dataset: dataset entity
        :param bool is_audio: is audio
        :return: annotation object
        :rtype: dtlpy.entities.annotation.Annotation
        """
        if item_metadata is None:
            item_metadata = dict()

        if is_video is None:
            if item is None:
                if _json['metadata'].get('system', {}).get('endFrame', 0) > 0:
                    is_video = True
                else:
                    is_video = False
            else:
                # get item type
                if item.mimetype is not None and 'video' in item.mimetype:
                    is_video = True

        if is_audio is None:
            if item is None:
                is_audio = False
            else:
                # get item type
                if item.mimetype is not None and 'audio' in item.mimetype:
                    is_audio = True

        item_url = _json.get('item', item.url if item is not None else None)
        item_id = _json.get('itemId', item.id if item is not None else None)
        dataset_url = _json.get('dataset', item.dataset_url if item is not None else None)
        dataset_id = _json.get('datasetId', item.dataset_id if item is not None else None)

        if item is not None:
            if item.id != item_id:
                logger.warning('Annotation has been fetched from a item that is not belong to it')
                item = None

        if dataset is not None:
            if dataset.id != dataset_id:
                logger.warning('Annotation has been fetched from a dataset that is not belong to it')
                dataset = None

        # get id
        annotation_id = None
        if 'id' in _json:
            annotation_id = _json['id']
        elif '_id' in _json:
            annotation_id = _json['_id']

        metadata = _json.get('metadata', dict())

        # get metadata, status, attributes and object id
        object_id = None
        status = None
        if 'system' in metadata and metadata['system'] is not None:
            object_id = _json['metadata']['system'].get('objectId', object_id)
            status = _json['metadata']['system'].get('status', status)

        named_attributes = metadata.get('system', dict()).get('attributes', None)
        recipe_1_attributes = _json.get('attributes', None)

        first_frame_attributes = recipe_1_attributes
        first_frame_coordinates = list()
        first_frame_number = 0
        first_frame_start_time = 0
        automated = None
        end_frame = None
        start_time = 0
        start_frame = 0
        end_time = None
        annotation_definition = None

        ############
        # if video #
        ############
        if is_video or is_audio:
            # get fps
            if item is not None and item.fps is not None:
                fps = item.fps
            if fps is None:
                if is_video:
                    fps = item_metadata.get('fps', 25)
                else:
                    item_metadata.get('fps', 1000)

            # get video-only attributes
            end_time = 1.5
            # get first frame attribute
            first_frame_attributes = first_frame_attributes
            # get first frame coordinates
            first_frame_coordinates = _json.get('coordinates', first_frame_coordinates)
            if 'system' in metadata:
                # get first frame number
                first_frame_number = _json['metadata']['system'].get('frame', first_frame_number)
                # get first frame start time
                start_time = _json['metadata']['system'].get('startTime', first_frame_start_time)
                # get first frame number
                start_frame = _json['metadata']['system'].get('frame', start_frame)
                automated = _json['metadata']['system'].get('automated', automated)
                end_frame = _json['metadata']['system'].get('endFrame', end_frame)
                end_time = _json['metadata']['system'].get('endTime', end_time)
        ################
        # if not video #
        ################
        if not is_video or is_audio:
            # get coordinates
            coordinates = _json.get('coordinates', list())
            # set video only attributes
            if not is_audio:
                end_time = 0
            # get automated
            if 'system' in metadata and metadata['system'] is not None:
                automated = metadata['system'].get('automated', automated)
            # set annotation definition
            def_dict = {'type': _json['type'],
                        'coordinates': coordinates,
                        'label': _json['label'],
                        'attributes': named_attributes}
            annotation_definition = FrameAnnotation.json_to_annotation_definition(def_dict)

        frames = entities.ReflectDict(
            value_type=FrameAnnotation,
            on_access=Annotation.on_access,
            start=start_frame,
            end=end_frame
        )

        # init annotation
        annotation = cls(
            # temp
            platform_dict=copy.deepcopy(_json),
            # platform
            id=annotation_id,
            url=_json.get('url', None),
            item_url=item_url,
            item=item,
            item_id=item_id,
            dataset=dataset,
            dataset_url=dataset_url,
            dataset_id=dataset_id,
            creator=_json['creator'],
            created_at=_json['createdAt'],
            updated_by=_json['updatedBy'],
            updated_at=_json['updatedAt'],
            hash=_json.get('hash', None),
            object_id=object_id,
            type=_json['type'],
            item_width=item_metadata.get('width', None),
            item_height=item_metadata.get('height', None),
            # meta
            metadata=metadata,
            fps=fps,
            status=status,
            # snapshots
            frames=frames,
            # video attributes
            automated=automated,
            end_frame=end_frame,
            end_time=end_time,
            start_frame=start_frame,
            annotations=annotations,
            start_time=start_time,
            label_suggestions=_json.get('labelSuggestions', None),
            source=_json.get('source', None),
            recipe_1_attributes=recipe_1_attributes,
        )
        annotation.annotation_definition = annotation_definition
        annotation.__client_api = client_api
        if annotation_definition:
            description = _json.get('description', None)
            if description is not None:
                annotation.description = description

        #################
        # if has frames #
        #################
        if is_video:
            # set first frame
            snapshot = {
                'attributes': first_frame_attributes,
                'coordinates': first_frame_coordinates,
                'fixed': True,
                'objectVisible': True,
                'frame': first_frame_number,
                'label': _json['label'],
                'type': annotation.type,
                'namedAttributes': named_attributes
            }

            # add first frame
            frame = FrameAnnotation.from_snapshot(
                _json=snapshot,
                annotation=annotation,
                fps=fps
            )
            annotation.frames[frame.frame_num] = frame
            annotation.annotation_definition = frame.annotation_definition

            # put snapshots if there are any
            for snapshot in _json['metadata']['system'].get('snapshots_', []):
                frame = FrameAnnotation.from_snapshot(
                    _json=snapshot,
                    annotation=annotation,
                    fps=fps
                )

                annotation.frames[frame.frame_num] = frame

            annotation.annotation_definition = annotation.frames[min(frames.actual_keys())].annotation_definition
            if annotation.annotation_definition:
                description = _json.get('description', None)
                if description is not None:
                    annotation.description = description

        return annotation

    @staticmethod
    def on_access(reflect_dict, actual_key: int, requested_key: int, val):
        val = copy.copy(val)
        val._interpolation = True
        val.fixed = False
        val.frame_num = requested_key
        reflect_dict[requested_key] = val
        return val

    def to_json(self):
        """
        Convert annotation object to a platform json representatio

        :return: platform json
        :rtype: dict
        """
        if len(self.frames.actual_keys()) > 0:
            self.set_frame(min(self.frames.actual_keys()))
        _json = attr.asdict(self,
                            filter=attr.filters.include(attr.fields(Annotation).id,
                                                        attr.fields(Annotation).url,
                                                        attr.fields(Annotation).metadata,
                                                        attr.fields(Annotation).creator,
                                                        attr.fields(Annotation).hash,
                                                        attr.fields(Annotation).metadata))

        # property attributes
        item_id = self.item_id
        if item_id is None and self._item is not None:
            item_id = self._item.id

        _json['itemId'] = item_id
        _json['item'] = self.item_url
        _json['label'] = self.label

        # Need to put back after transition to attributes 2.0
        # _json['attributes'] = self.attributes

        _json['dataset'] = self.dataset_url

        _json['createdAt'] = self.created_at
        _json['updatedBy'] = self.updated_by
        _json['updatedAt'] = self.updated_at
        _json['source'] = self.source

        if self.description is not None:
            _json['description'] = self.description

        if self.label_suggestions:
            _json['labelSuggestions'] = self.label_suggestions

        if self._item is not None and self.dataset_id is None:
            _json['datasetId'] = self._item.dataset_id
        else:
            _json['datasetId'] = self.dataset_id

        _json['type'] = self.type
        if self.type != 'class':
            _json['coordinates'] = self.coordinates

        # add system metadata
        if _json['metadata'].get('system', None) is None:
            _json['metadata']['system'] = dict()
        if self.automated is not None:
            _json['metadata']['system']['automated'] = self.automated
        if self.object_id is not None:
            _json['metadata']['system']['objectId'] = self.object_id
        if self.status is not None:
            # if status is CLEAR need to set status to None so it will be deleted in backend
            _json['metadata']['system']['status'] = self.status if self.status != AnnotationStatus.CLEAR else None

        if isinstance(self.annotation_definition, entities.Description):
            _json['metadata']['system']['system'] = True

        _json['metadata']['system']['attributes'] = self.attributes if self.attributes is not None else dict()
        _json['attributes'] = self._recipe_1_attributes


        # add frame info
        if self.is_video or (self.end_time and self.end_time > 0) or (self.end_frame and self.end_frame > 0):
            # get all snapshots but the first one
            snapshots = list()
            frame_numbers = self.frames.actual_keys() if len(self.frames.actual_keys()) else []
            for frame_num in sorted(frame_numbers):
                if frame_num <= self.frames.start:
                    continue
                if frame_num > self.frames.end:
                    break
                if not self.frames[frame_num]._interpolation or self.frames[frame_num].fixed:
                    snapshots.append(self.frames[frame_num].to_snapshot())
                    self.frames[frame_num]._interpolation = False
            # add metadata to json
            _json['metadata']['system']['frame'] = self.start_frame
            _json['metadata']['system']['startTime'] = self.start_time
            _json['metadata']['system']['endTime'] = self.end_time
            if self.end_frame is not None:
                _json['metadata']['system']['endFrame'] = self.end_frame

            # add snapshots only if classification
            if self.type not in ['subtitle']:
                _json['metadata']['system']['snapshots_'] = snapshots

        return _json

    def task_scores(self, task_id: str, page_offset: int = None, page_size: int = None):
        """
        Get the scores of the annotation in a specific task.
        :param task_id: The ID of the task.
        :param page_offset: The page offset.
        :param page_size: The page size.
        :return: page of scores
        """
        return self.annotations.task_scores(annotation_id=self.id ,task_id=task_id, page_offset=page_offset, page_size=page_size)



@attr.s
class FrameAnnotation(entities.BaseEntity):
    """
    FrameAnnotation object
    """
    # parent annotation
    annotation = attr.ib()

    # annotations
    annotation_definition = attr.ib()

    # multi
    frame_num = attr.ib()
    fixed = attr.ib()
    object_visible = attr.ib()

    # temp
    _interpolation = attr.ib(repr=False, default=False)
    _recipe_1_attributes = attr.ib(repr=False, default=None)

    ################################
    # parent annotation attributes #
    ################################

    @property
    def status(self):
        return self.annotation.status

    @property
    def timestamp(self):
        if self.annotation.fps is not None and self.frame_num is not None:
            return self.frame_num / self.annotation.fps if self.annotation.fps != 0 else None

    ####################################
    # annotation definition attributes #
    ####################################
    @property
    def type(self):
        return self.annotation.type

    @property
    def label(self):
        return self.annotation_definition.label

    @label.setter
    def label(self, label):
        self.annotation_definition.label = label

    @property
    def attributes(self):
        return self.annotation_definition.attributes

    @attributes.setter
    def attributes(self, attributes):
        self.annotation_definition.attributes = attributes

    @property
    def geo(self):
        return self.annotation_definition.geo

    @property
    def top(self):
        return self.annotation_definition.top

    @property
    def bottom(self):
        return self.annotation_definition.bottom

    @property
    def left(self):
        return self.annotation_definition.left

    @property
    def right(self):
        return self.annotation_definition.right

    @property
    def color(self):
        if self.annotation.item is None:
            return 255, 255, 255
        else:
            label = None
            for label in self.annotation.item.dataset.labels:
                if label.tag.lower() == self.label.lower():
                    return label.rgb
            if label is None:
                return 255, 255, 255

    @property
    def coordinates(self):
        coordinates = self.annotation_definition.to_coordinates(color=self.color)
        return coordinates

    @property
    def x(self):
        return self.annotation_definition.x

    @property
    def y(self):
        return self.annotation_definition.y

    @property
    def rx(self):
        if self.annotation_definition.type == 'ellipse':
            return self.annotation_definition.rx
        else:
            return None

    @property
    def ry(self):
        if self.annotation_definition.type == 'ellipse':
            return self.annotation_definition.ry
        else:
            return None

    @property
    def angle(self):
        if self.annotation_definition.type in ['ellipse', 'box']:
            return self.annotation_definition.angle
        else:
            return None

    ######################
    # annotation methods #
    ######################
    def show(self, **kwargs):
        """
        Show annotation as ndarray
        :param kwargs: see annotation definition
        :return: ndarray of the annotation
        """
        return self.annotation_definition.show(**kwargs)

    @staticmethod
    def json_to_annotation_definition(_json):
        if 'namedAttributes' in _json:
            _json['attributes'] = _json['namedAttributes']
        else:
            if not isinstance(_json.get('attributes'), dict):
                _json['attributes'] = None
        if _json['type'] == 'segment':
            annotation = entities.Polygon.from_json(_json)
        elif _json['type'] == 'polyline':
            annotation = entities.Polyline.from_json(_json)
        elif _json['type'] == 'box':
            annotation = entities.Box.from_json(_json)
        elif _json['type'] == 'cube':
            annotation = entities.Cube.from_json(_json)
        elif _json['type'] == 'cube_3d':
            annotation = entities.Cube3d.from_json(_json)
        elif _json['type'] == 'point':
            annotation = entities.Point.from_json(_json)
        elif _json['type'] == 'binary':
            annotation = entities.Segmentation.from_json(_json)
        elif _json['type'] == 'class':
            annotation = entities.Classification.from_json(_json)
        elif _json['type'] == 'subtitle':
            annotation = entities.Subtitle.from_json(_json)
        elif _json['type'] == 'ellipse':
            annotation = entities.Ellipse.from_json(_json)
        elif _json['type'] == 'comparison':
            annotation = entities.Comparison.from_json(_json)
        elif _json['type'] == 'note':
            annotation = entities.Note.from_json(_json)
        elif _json['type'] == 'pose':
            annotation = entities.Pose.from_json(_json)
        elif _json['type'] == 'gis':
            annotation = entities.Gis.from_json(_json)
        else:
            annotation = entities.UndefinedAnnotationType.from_json(_json)
        return annotation

    #######
    # I/O #
    #######
    @classmethod
    def new(cls, annotation, annotation_definition, frame_num, fixed, object_visible=True):
        """
        new frame state to annotation

        :param annotation: annotation
        :param annotation_definition: annotation type object - must be same type as annotation
        :param frame_num: frame number
        :param fixed: is fixed
        :param object_visible: does the annotated object is visible
        :return: FrameAnnotation object
        """
        frame = cls(
            # annotations
            annotation=annotation,
            annotation_definition=annotation_definition,

            # multi
            frame_num=frame_num,
            fixed=fixed,
            object_visible=object_visible,
        )
        if annotation_definition.attributes:
            frame.attributes = annotation_definition.attributes
        return frame

    @classmethod
    def from_snapshot(cls, annotation, _json, fps=None):
        """
        new frame state to annotation

        :param annotation: annotation
        :param _json: annotation type object - must be same type as annotation
        :param fps: frame number
        :return: FrameAnnotation object
        """
        # get annotation class
        _json['type'] = annotation.type
        attrs = _json.get('attributes', None)
        annotation_definition = cls.json_to_annotation_definition(_json=_json)

        frame_num = _json.get('frame', annotation.last_frame + 1)

        return cls(
            # annotations
            annotation=annotation,
            annotation_definition=annotation_definition,

            # multi
            frame_num=frame_num,
            fixed=_json.get('fixed', False),
            object_visible=_json.get('objectVisible', True),
            recipe_1_attributes=attrs,
        )

    def to_snapshot(self):

        snapshot_dict = {
            'frame': self.frame_num,
            'fixed': self.fixed,
            'label': self.label,
            'type': self.type,
            'objectVisible': self.object_visible,
            'data': self.coordinates
        }

        if self.annotation_definition.description is not None:
            snapshot_dict['description'] = self.annotation_definition.description

        if self.attributes is not None:
            snapshot_dict['namedAttributes'] = self.attributes

        if self._recipe_1_attributes is not None:
            snapshot_dict['attributes'] = self._recipe_1_attributes


        return snapshot_dict



from typing import List
import logging

from .. import entities, exceptions
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class UnsearchablePaths:
    """
    Unsearchable Paths

    """

    def __init__(self, client_api: ApiClient, dataset: entities.Dataset = None):
        self._client_api = client_api
        self._dataset = dataset

    @property
    def dataset(self) -> entities.Dataset:
        if self._dataset is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Cannot perform action WITHOUT Dataset entity in {} repository.'.format(
                    self.__class__.__name__) + ' Please use dataset.schema or set a dataset')
        assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    def __unsearchable_paths_request(self, payload):
        """
        Set unsearchable paths in dataset schema
        """
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/datasets/{}/schema/items'.format(self.dataset.id),
                                                         json_req=
                                                         {
                                                             "unsearchablePaths": payload
                                                         })
        if not success:
            raise exceptions.PlatformException(response)

        resp = response.json()
        if isinstance(resp, dict):
            command = entities.Command.from_json(_json=resp,
                                                 client_api=self._client_api)

            try:
                command.wait()
            except Exception as e:
                logger.error('Command failed: {}'.format(e))
        else:
            logger.warning(resp)
        return success

    def add(self, paths: List[str]):
        """
        Add metadata paths to `unsearchablePaths` to exclude keys under these paths from indexing, making them unsearchable through the Dataset Browser UI and DQL queries.

        :param paths: list of paths to create
        :return: true if success, else raise exception
        :rtype: bool

        **Example**:

        .. code-block:: python

            success = dataset.schema.unsearchable_paths.add(paths=['metadata.key1', 'metadata.key2'])
        """
        return self.__unsearchable_paths_request(payload={"add": paths})

    def remove(self, paths: List[str]):
        """
        Remove metadata paths from `unsearchablePaths` to index keys under these paths, making them searchable through the Dataset Browser UI and DQL queries.

        :param paths: list of paths to delete
        :return: true if success, else raise exception
        :rtype: bool

        **Example**:

        .. code-block:: python

            success = dataset.schema.unsearchable_paths.remove(paths=['metadata.key1', 'metadata.key2'])
        """
        return self.__unsearchable_paths_request(payload={"remove": paths})


class Schema:
    """
    Schema Repository
    """

    def __init__(self, client_api: ApiClient, dataset: entities.Dataset):
        self._client_api = client_api
        self.dataset = dataset
        self.unsearchable_paths = UnsearchablePaths(client_api=self._client_api, dataset=dataset)

    ###########
    # methods #
    ###########
    def get(self):
        """
        Get dataset schema

        :return: dataset schema
        :rtype: dict

        **Example**:

        .. code-block:: python

            json = dataset.schema.get()
        """
        success, response = self._client_api.gen_request(req_type='get',
                                                         path='/datasets/{}/schema'.format(self.dataset.id))
        if not success:
            raise exceptions.PlatformException(response)

        return response.json()



import dtlpy as dl
import logging

logger = logging.getLogger(name='dtlpy')


class ServiceRunner(dl.BaseServiceRunner):
    """
    Package runner class

    """

    def __init__(self):
        """
        Init package attributes here

        :return:
        """

    def run(self, item, dataset, config, annotation, progress=None):
        """
        In this example we upload annotation to item, update its metadata and add it to a task in the dataset

        :param progress: Use this to update the progress of your package
        :return:
        """
        # these lines can be removed
        assert isinstance(progress, dl.Progress)
        assert isinstance(item, dl.Item)
        assert isinstance(dataset, dl.entities.Dataset)
        assert isinstance(annotation, dl.Annotation)
        item.annotations.upload(annotations=annotation)
        item.metadata['user'] = config['metadata']
        item.update()
        if annotation.label == 'street':
            task = dataset.tasks.get(task_name='streetAnnotation')
            task.add_items(items=item)
        logger.info('Function completed successfully')


if __name__ == "__main__":
    """
    Run this main to locally debug your package
    """
    dl.packages.test_local_package()



import attr
import traceback
import logging
from enum import Enum
from collections import namedtuple

from .. import entities, exceptions, repositories
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class TriggerResource(str, Enum):
    ITEM = "Item"
    DATASET = "Dataset"
    ANNOTATION = "Annotation"
    TASK = 'Task',
    ASSIGNMENT = 'Assignment',
    ITEM_STATUS = "ItemStatus"


class TriggerAction(str, Enum):
    CREATED = "Created"
    UPDATED = "Updated"
    DELETED = "Deleted"
    STATUS_CHANGED = 'statusChanged'
    CLONE = 'Clone'


class TriggerExecutionMode(str, Enum):
    ONCE = "Once"
    ALWAYS = "Always"


class TriggerType(str, Enum):
    EVENT = "Event"
    CRON = "Cron"


@attr.s
class BaseTrigger(entities.BaseEntity):
    """
    Trigger Entity
    """
    #######################
    # Platform attributes #
    #######################
    id = attr.ib()
    url = attr.ib(repr=False)
    created_at = attr.ib()
    updated_at = attr.ib(repr=False)
    creator = attr.ib()
    name = attr.ib()
    active = attr.ib()
    type = attr.ib()
    scope = attr.ib()
    is_global = attr.ib()
    input = attr.ib()

    # name change
    function_name = attr.ib()
    service_id = attr.ib()
    webhook_id = attr.ib()
    pipeline_id = attr.ib()

    ########
    # temp #
    ########
    special = attr.ib(repr=False)

    ##############################
    # different name in platform #
    ##############################
    project_id = attr.ib()
    _spec = attr.ib()
    operation = attr.ib()

    ##################
    # SDK attributes #
    ##################
    _service = attr.ib(repr=False)
    _project = attr.ib(repr=False)
    _client_api = attr.ib(type=ApiClient, repr=False)
    _op_type = attr.ib(default='service')
    _repositories = attr.ib(repr=False)

    updated_by = attr.ib(default=None)

    @staticmethod
    def _get_operation(operation):
        op_type = operation.get('type', None)
        if op_type == 'function':
            service_id = operation.get('serviceId', None)
            webhook_id = None
            pipeline_id = None
        elif op_type == 'webhook':
            webhook_id = operation.get('webhookId', None)
            service_id = None
            pipeline_id = None
        elif op_type == 'rabbitmq':
            webhook_id = None
            service_id = None
            pipeline_id = None
        elif op_type == 'pipeline':
            webhook_id = None
            service_id = None
            pipeline_id = operation.get('id', None)
        else:
            raise exceptions.PlatformException('400', 'unknown trigger operation type: {}'.format(op_type))

        return service_id, webhook_id, pipeline_id

    @property
    def createdAt(self):
        return self.created_at

    @property
    def updatedAt(self):
        return self.updated_at

    @staticmethod
    def _protected_from_json(_json, client_api, project=None, service=None):
        """
        Same as from_json but with try-except to catch if error

        :param _json: platform json
        :param client_api: ApiClient entity
        :param project: project entity
        :param service: service entity
        :return:
        """
        try:
            trigger = BaseTrigger.from_json(_json=_json,
                                            client_api=client_api,
                                            project=project,
                                            service=service)
            status = True
        except Exception:
            trigger = traceback.format_exc()
            status = False
        return status, trigger

    @classmethod
    def from_json(cls, _json, client_api, project=None, service=None):
        """
        Build a trigger entity object from a json

        :param dict _json: platform json
        :param dl.ApiClient client_api: ApiClient entity
        :param dtlpy.entities.project.Project project: project entity
        :param dtlpy.entities.service.Service service: service entity
        :return:
        """
        if project is not None:
            if project.id != _json.get('projectId', None):
                logger.warning('Trigger has been fetched from a project that is not belong to it')
                project = None

        if service is not None:
            spec = _json.get('spec', dict())
            operation = spec.get('operation', dict())
            if service.id != operation.get('serviceId', None):
                logger.warning('Trigger has been fetched from a service that is not belong to it')
                service = None

        trigger_type = _json.get('type', None)

        if trigger_type == TriggerType.CRON:
            ent = CronTrigger.from_json(_json, client_api, project, service)
        else:
            ent = Trigger.from_json(_json, client_api, project, service)
        return ent

    ################
    # repositories #
    ################
    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['services', 'triggers'])

        if self._project is None:
            services_repo = repositories.Services(client_api=self._client_api, project=self._project)
        else:
            services_repo = self._project.services

        triggers = repositories.Triggers(client_api=self._client_api,
                                         project=self._project)

        r = reps(services=services_repo, triggers=triggers)
        return r

    @property
    def triggers(self):
        assert isinstance(self._repositories.triggers, repositories.Triggers)
        return self._repositories.triggers

    @property
    def services(self):
        assert isinstance(self._repositories.services, repositories.Services)
        return self._repositories.services

    ############
    # entities #
    ############
    @property
    def project(self):
        if self._project is None:
            self._project = repositories.Projects(client_api=self._client_api).get(project_id=self.project_id,
                                                                                   fetch=None)
        assert isinstance(self._project, entities.Project)
        return self._project

    @property
    def service(self):
        if self._service is None:
            self._service = self.services.get(service_id=self.service_id, fetch=None)
        assert isinstance(self._service, entities.Service)
        return self._service

    ###########
    # methods #
    ###########
    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """
        # get excluded
        _json = attr.asdict(self, filter=attr.filters.exclude(attr.fields(BaseTrigger)._client_api,
                                                              attr.fields(BaseTrigger).project_id,
                                                              attr.fields(BaseTrigger)._project,
                                                              attr.fields(BaseTrigger)._service,
                                                              attr.fields(BaseTrigger).special,
                                                              attr.fields(BaseTrigger)._op_type,
                                                              attr.fields(BaseTrigger)._spec,
                                                              attr.fields(BaseTrigger)._repositories,
                                                              attr.fields(BaseTrigger).service_id,
                                                              attr.fields(BaseTrigger).webhook_id,
                                                              attr.fields(BaseTrigger).pipeline_id,
                                                              attr.fields(BaseTrigger).function_name,
                                                              attr.fields(BaseTrigger).is_global,
                                                              attr.fields(BaseTrigger).created_at,
                                                              attr.fields(BaseTrigger).updated_at,
                                                              attr.fields(BaseTrigger).operation,
                                                              attr.fields(BaseTrigger).updated_by,
                                                              ))

        # rename
        _json['projectId'] = self.project_id
        _json['createdAt'] = self.created_at
        _json['updatedAt'] = self.updated_at
        if self.is_global is not None:
            _json['global'] = self.is_global
        if self.updated_by is not None:
            _json['updatedBy'] = self.updated_by
        return _json

    def delete(self):
        """
        Delete Trigger object

        :return: True
        """
        return self.project.triggers.delete(trigger_id=self.id)

    def update(self):
        """
        Update Trigger object

        :return: Trigger entity
        """
        return self.project.triggers.update(trigger=self)


@attr.s
class Trigger(BaseTrigger):
    """
    Trigger Entity
    """
    filters = attr.ib(default=None, repr=False)
    execution_mode = attr.ib(default=TriggerExecutionMode.ONCE, repr=False)
    actions = attr.ib(default=TriggerAction.CREATED, repr=False)
    resource = attr.ib(default=TriggerResource.ITEM, repr=False)

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """
        _json = super().to_json()

        _json['spec'] = {
            'filter': _json.pop('filters'),
            'executionMode': _json.pop('execution_mode'),
            'resource': _json.pop('resource'),
            'actions': _json.pop('actions'),
            'input': _json.pop('input', None),
            'operation': self.operation,
        }
        return _json

    @classmethod
    def from_json(cls, _json, client_api, project=None, service=None):
        """
        Build a trigger entity object from a json

        :param _json: platform json
        :param client_api: ApiClient entity
        :param dtlpy.entities.project.Project project: project entity
        :param dtlpy.entities.service.Service service: service entity
        :return:
        """
        spec = _json.get('spec', dict())
        operation = spec.get('operation', dict())

        service_id, webhook_id, pipeline_id = cls._get_operation(operation=operation)

        return cls(
            execution_mode=spec.get('executionMode', None),
            updated_at=_json.get('updatedAt', None),
            created_at=_json.get('createdAt', None),
            resource=spec.get('resource', None),
            creator=_json.get('creator', None),
            special=_json.get('special', None),
            actions=spec.get('actions', None),
            active=_json.get('active', None),
            function_name=operation.get('functionName', None),
            scope=_json.get('scope', None),
            is_global=_json.get('global', None),
            type=_json.get('type', None),
            name=_json.get('name', None),
            url=_json.get('url', None),
            service_id=service_id,
            project_id=_json.get('projectId', None),
            input=spec.get('input', None),
            webhook_id=webhook_id,
            client_api=client_api,
            filters=spec.get('filter', dict()),
            project=project,
            service=service,
            id=_json.get('id', None),
            op_type=operation.get('type', None),
            spec=spec,
            pipeline_id=pipeline_id,
            operation=operation,
            updated_by=_json.get('updatedBy', None),
        )


@attr.s
class CronTrigger(BaseTrigger):
    start_at = attr.ib(default=None)
    end_at = attr.ib(default=None)
    cron = attr.ib(default=None)

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """
        _json = super().to_json()

        _json['spec'] = {
            'startAt': _json.pop('start_at'),
            'endAt': _json.pop('end_at'),
            'cron': _json.pop('cron'),
            'input': _json.pop('input'),
            'operation': self.operation,
        }
        return _json

    @classmethod
    def from_json(cls, _json, client_api, project=None, service=None):
        """
        Build a trigger entity object from a json

        :param _json: platform json
        :param client_api: ApiClient entity
        :param project: project entity
        :param service: service entity
        :return:
        """
        spec = _json.get('spec', dict())
        operation = spec.get('operation', dict())

        project_id = _json.get('projectId', None)
        if project_id is not None and project is not None:
            if project_id != project.id:
                project = None

        service_id, webhook_id, pipeline_id = cls._get_operation(operation=operation)
        return cls(
            updated_at=_json.get('updatedAt', None),
            created_at=_json.get('createdAt', None),
            creator=_json.get('creator', None),
            special=_json.get('special', None),
            active=_json.get('active', None),
            function_name=operation.get('functionName', None),
            scope=_json.get('scope', None),
            is_global=_json.get('global', None),
            type=_json.get('type', None),
            name=_json.get('name', None),
            input=spec.get('input', None),
            end_at=spec.get('endAt', None),
            start_at=spec.get('startAt', None),
            cron=spec.get('cron', None),
            url=_json.get('url', None),
            service_id=service_id,
            project_id=project_id,
            webhook_id=webhook_id,
            client_api=client_api,
            project=project,
            service=service,
            id=_json['id'],
            op_type=operation.get('type', None),
            spec=spec,
            pipeline_id=pipeline_id,
            operation=operation
        )



from collections import namedtuple
from enum import Enum
import traceback
import mimetypes
import logging
import attr
import copy
import os
import io
from .. import repositories, entities, exceptions
from .annotation import ViewAnnotationOptions, ExportVersion
from ..services.api_client import ApiClient
from ..services.api_client import client as client_api
import json
from typing import List
import requests

logger = logging.getLogger(name='dtlpy')


class ExportMetadata(Enum):
    FROM_JSON = 'from_json'


class ItemStatus(str, Enum):
    COMPLETED = "completed"
    APPROVED = "approved"
    DISCARDED = "discard"


@attr.s
class Item(entities.BaseEntity):
    """
    Item object
    """
    # item information
    annotations_link = attr.ib(repr=False)
    dataset_url = attr.ib()
    thumbnail = attr.ib(repr=False)
    created_at = attr.ib()
    updated_at = attr.ib()
    updated_by = attr.ib()
    dataset_id = attr.ib()
    annotated = attr.ib(repr=False)
    metadata = attr.ib(repr=False)
    filename = attr.ib()
    stream = attr.ib(repr=False)
    name = attr.ib()
    type = attr.ib()
    url = attr.ib(repr=False)
    id = attr.ib()
    hidden = attr.ib(repr=False)
    dir = attr.ib(repr=False)
    spec = attr.ib()
    creator = attr.ib()
    _description = attr.ib()
    _src_item = attr.ib(repr=False)

    # name change
    annotations_count = attr.ib()

    # api
    _client_api = attr.ib(type=ApiClient, repr=False)
    _platform_dict = attr.ib(repr=False)

    # entities
    _dataset = attr.ib(repr=False)
    _model = attr.ib(repr=False)
    _project = attr.ib(repr=False)
    _project_id = attr.ib(repr=False)

    # repositories
    _repositories = attr.ib(repr=False)

    @property
    def createdAt(self):
        return self.created_at

    @property
    def datasetId(self):
        return self.dataset_id

    @property
    def resolved_stream(self):
        stream = self.metadata.get('system', dict()).get('shebang', dict()).get('linkInfo', dict()).get('ref', None)
        if stream is None:
            stream = self.stream
            api_url = self._client_api.environment
            if api_url != self._client_api.base_gate_url:
                stream = stream.replace(api_url, self._client_api.base_gate_url)
        else:
            link_item_url_override = os.environ.get('LINK_ITEM_URL_OVERRIDE', None)
            if link_item_url_override is not None:
                src, target = link_item_url_override.split(',')
                stream = stream.replace(src, target)

        return stream

    @staticmethod
    def _protected_from_json(_json, client_api, dataset=None):
        """
        Same as from_json but with try-except to catch if error
        :param _json: platform json
        :param client_api: ApiClient entity
        :param dataset: dataset entity
        :return:
        """
        try:
            item = Item.from_json(_json=_json,
                                  client_api=client_api,
                                  dataset=dataset)
            status = True
        except Exception:
            item = traceback.format_exc()
            status = False
        return status, item

    @classmethod
    def from_json(cls, _json, client_api, dataset=None, project=None, model=None, is_fetched=True):
        """
        Build an item entity object from a json

        :param dtlpy.entities.project.Project project: project entity
        :param dict _json: _json response from host
        :param dtlpy.entities.dataset.Dataset dataset: dataset in which the annotation's item is located
        :param dtlpy.entities.dataset.Model model: the model entity if item is an artifact of a model
        :param dlApiClient client_api: ApiClient entity
        :param bool is_fetched: is Entity fetched from Platform
        :return: Item object
        :rtype: dtlpy.entities.item.Item
        """
        dataset_id = None
        if dataset is not None:
            dataset_id = _json.get('datasetId', None)
            if dataset.id != dataset_id and dataset_id is not None:
                logger.warning('Item has been fetched from a dataset that is not belong to it')
                dataset = None
            else:
                dataset_id = dataset.id

        metadata = _json.get('metadata', dict())
        project_id = _json.get('projectId', None)
        if project_id is None:
            project_id = project.id if project else None
        inst = cls(
            # sdk
            platform_dict=copy.deepcopy(_json),
            client_api=client_api,
            dataset=dataset,
            project=project,
            model=model,
            # params
            annotations_link=_json.get('annotations', None),
            thumbnail=_json.get('thumbnail', None),
            dataset_id=_json.get('datasetId', dataset_id),
            annotated=_json.get('annotated', None),
            dataset_url=_json.get('dataset', None),
            created_at=_json.get('createdAt', None),
            annotations_count=_json.get('annotationsCount', None),
            hidden=_json.get('hidden', False),
            stream=_json.get('stream', None),
            dir=_json.get('dir', None),
            filename=_json.get('filename', None),
            metadata=metadata,
            name=_json.get('name', None),
            type=_json.get('type', None),
            url=_json.get('url', None),
            id=_json.get('id', None),
            spec=_json.get('spec', None),
            creator=_json.get('creator', None),
            project_id=project_id,
            description=_json.get('description', None),
            src_item=_json.get('srcItem', None),
            updated_at=_json.get('updatedAt', None),
            updated_by=_json.get('updatedBy', None)
        )
        inst.is_fetched = is_fetched
        return inst

    def __getstate__(self):
        # dump state to json
        return self.to_json()

    def __setstate__(self, state):
        # create a new item, and update the current one with the same state
        # this way we can have _client_api, and all the repositories and entities which are not picklable
        self.__dict__.update(entities.Item.from_json(_json=state,
                                                     client_api=client_api).__dict__)

    ############
    # entities #
    ############
    @property
    def dataset(self):
        if self._dataset is None:
            self._dataset = self.datasets.get(dataset_id=self.dataset_id, fetch=None)
        assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    @property
    def model(self):
        return self._model

    def __update_item_binary(self, _json):
        binary = io.BytesIO()
        binary.write(json.dumps(_json).encode())
        binary.seek(0)
        binary.name = self.name
        success, resp = client_api.gen_request(req_type='post',
                                      path=f'/items/{self.id}/revisions',
                                      files={'file': (binary.name, binary)})
        if not success:
            raise exceptions.PlatformException(resp)

    @property
    def project(self):
        if self._project is None:
            if self._dataset is None:
                self._dataset = self.datasets.get(dataset_id=self.dataset_id, fetch=None)
            self._project = self._dataset.project
            if self._project is None:
                raise exceptions.PlatformException(error='2001',
                                                   message='Missing entity "project".')
        assert isinstance(self._project, entities.Project)
        return self._project

    @property
    def project_id(self):
        if self._project_id is None:
            if self._dataset is None:
                self._dataset = self.datasets.get(dataset_id=self.dataset_id, fetch=None)
            self._project_id = self._dataset.project.id
        return self._project_id

    ################
    # repositories #
    ################

    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['annotations', 'datasets', 'items', 'codebases', 'artifacts', 'modalities',
                                       'features', 'assignments', 'tasks', 'resource_executions', 'collections'])
        reps.__new__.__defaults__ = (None, None, None, None, None, None, None, None, None)

        if self._dataset is None:
            items = repositories.Items(
                client_api=self._client_api,
                dataset=self._dataset,
                dataset_id=self.dataset_id,
                datasets=repositories.Datasets(client_api=self._client_api, project=None)
            )
            datasets = items.datasets

        else:
            items = self.dataset.items
            datasets = self.dataset.datasets

        r = reps(
            annotations=repositories.Annotations(
                client_api=self._client_api,
                dataset_id=self.dataset_id,
                item=self,
                dataset=self._dataset
            ),
            items=items,
            datasets=datasets,
            codebases=None,
            artifacts=None,
            modalities=Modalities(item=self),
            features=repositories.Features(
                client_api=self._client_api,
                project=self._project,
                item=self
            ),
            tasks=repositories.Tasks(
                client_api=self._client_api,
                project=self._project,
                dataset=self._dataset
            ),
            assignments=repositories.Assignments(
                client_api=self._client_api,
                project=self._project,
                dataset=self._dataset
            ),
            resource_executions=repositories.ResourceExecutions(
                client_api=self._client_api,
                project=self._project,
                resource=self
            ),
            collections=repositories.Collections(client_api=self._client_api, item=self, dataset=self._dataset)
        )
        return r

    @property
    def modalities(self):
        assert isinstance(self._repositories.modalities, Modalities)
        return self._repositories.modalities

    @property
    def annotations(self):
        assert isinstance(self._repositories.annotations, repositories.Annotations)
        return self._repositories.annotations

    @property
    def datasets(self):
        assert isinstance(self._repositories.datasets, repositories.Datasets)
        return self._repositories.datasets

    @property
    def assignments(self):
        assert isinstance(self._repositories.assignments, repositories.Assignments)
        return self._repositories.assignments

    @property
    def tasks(self):
        assert isinstance(self._repositories.tasks, repositories.Tasks)
        return self._repositories.tasks

    @property
    def resource_executions(self):
        assert isinstance(self._repositories.resource_executions, repositories.ResourceExecutions)
        return self._repositories.resource_executions

    @property
    def items(self):
        assert isinstance(self._repositories.items, repositories.Items)
        return self._repositories.items

    @property
    def features(self):
        assert isinstance(self._repositories.features, repositories.Features)
        return self._repositories.features
    
    @property
    def collections(self):
        assert isinstance(self._repositories.collections, repositories.Collections)
        return self._repositories.collections

    ##############
    # Properties #
    ##############
    @property
    def height(self):
        return self.metadata.get('system', dict()).get('height', None)

    @height.setter
    def height(self, val):
        if 'system' not in self.metadata:
            self.metadata['system'] = dict()
        self.metadata['system']['height'] = val

    @property
    def width(self):
        return self.metadata.get('system', dict()).get('width', None)

    @width.setter
    def width(self, val):
        if 'system' not in self.metadata:
            self.metadata['system'] = dict()
        self.metadata['system']['width'] = val

    @property
    def fps(self):
        return self.metadata.get('fps', None)

    @fps.setter
    def fps(self, val):
        self.metadata['fps'] = val

    @property
    def mimetype(self):
        mimetype = self.metadata.get('system', dict()).get('mimetype', None)
        if mimetype is None:
            mimetype = mimetypes.guess_type(self.filename)[0]
        return mimetype

    @property
    def size(self):
        return self.metadata.get('system', dict()).get('size', None)

    @property
    def system(self):
        return self.metadata.get('system', dict())

    @property
    def description(self):
        description = None
        if self._description is not None:
            description = self._description
        elif 'description' in self.metadata:
            description = self.metadata['description'].get('text', None)
        return description

    @property
    def platform_url(self):
        return self._client_api._get_resource_url(
            "projects/{}/datasets/{}/items/{}".format(self.dataset.projects[-1], self.dataset.id, self.id))

    @description.setter
    def description(self, text: str):
        """
        Update Item description

        :param text: if None or "" description will be deleted
        :return
        """
        self.set_description(text=text)

    ###########
    # Functions #
    ###########
    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """
        _json = attr.asdict(self,
                            filter=attr.filters.exclude(attr.fields(Item)._repositories,
                                                        attr.fields(Item)._dataset,
                                                        attr.fields(Item)._model,
                                                        attr.fields(Item)._project,
                                                        attr.fields(Item)._client_api,
                                                        attr.fields(Item)._platform_dict,
                                                        attr.fields(Item).annotations_count,
                                                        attr.fields(Item).dataset_url,
                                                        attr.fields(Item).annotations_link,
                                                        attr.fields(Item).spec,
                                                        attr.fields(Item).creator,
                                                        attr.fields(Item).created_at,
                                                        attr.fields(Item).dataset_id,
                                                        attr.fields(Item)._project_id,
                                                        attr.fields(Item)._description,
                                                        attr.fields(Item)._src_item,
                                                        attr.fields(Item).updated_at,
                                                        attr.fields(Item).updated_by
                                                        ))

        _json.update({'annotations': self.annotations_link,
                      'annotationsCount': self.annotations_count,
                      'dataset': self.dataset_url,
                      'createdAt': self.created_at,
                      'datasetId': self.dataset_id,
                      })
        if self.spec is not None:
            _json['spec'] = self.spec
        if self.creator is not None:
            _json['creator'] = self.creator
        if self._description is not None:
            _json['description'] = self.description
        if self._src_item is not None:
            _json['srcItem'] = self._src_item

        _json['updatedAt'] = self.updated_at
        _json['updatedBy'] = self.updated_by

        return _json

    def download(
            self,
            # download options
            local_path=None,
            file_types=None,
            save_locally=True,
            to_array=False,
            annotation_options: ViewAnnotationOptions = None,
            overwrite=False,
            to_items_folder=True,
            thickness=1,
            with_text=False,
            annotation_filters=None,
            alpha=1,
            export_version=ExportVersion.V1,
            dataset_lock=False,
            lock_timeout_sec=None,
            export_summary=False,
            raise_on_error=False
    ):
        """
        Download dataset by filters.
        Filtering the dataset for items and save them local
        Optional - also download annotation, mask, instance and image mask of the item

        :param str local_path: local folder or filename to save to.
        :param list file_types: a list of file type to download. e.g ['video/webm', 'video/mp4', 'image/jpeg', 'image/png']
        :param bool save_locally: bool. save to disk or return a buffer
        :param bool to_array: returns Ndarray when True and local_path = False
        :param list annotation_options: download annotations options:  list(dl.ViewAnnotationOptions)
        :param dtlpy.entities.filters.Filters annotation_filters: Filters entity to filter annotations for download
        :param bool overwrite: optional - default = False
        :param bool dataset_lock: optional - default = False
        :param bool export_summary: optional - default = False
        :param int lock_timeout_sec: optional
        :param bool to_items_folder: Create 'items' folder and download items to it
        :param int thickness: optional - line thickness, if -1 annotation will be filled, default =1
        :param bool with_text: optional - add text to annotations, default = False
        :param float alpha: opacity value [0 1], default 1
        :param str export_version:  exported items will have original extension in filename, `V1` - no original extension in filenames
        :param bool raise_on_error: raise an exception if an error occurs
        :return: generator of local_path per each downloaded item
        :rtype: generator or single item

        **Example**:

        .. code-block:: python

            item.download(local_path='local_path',
                         annotation_options=dl.ViewAnnotationOptions.MASK,
                         overwrite=False,
                         thickness=1,
                         with_text=False,
                         alpha=1,
                         save_locally=True,
                         dataset_lock=False
                         lock_timeout_sec=300,
                         export_summary=False
                         )
        """
        # if dir - concatenate local path and item name
        if local_path is not None:
            if os.path.isdir(local_path):
                local_path = os.path.join(local_path, self.name)
            else:
                _, ext = os.path.splitext(local_path)
                _, item_ext = os.path.splitext(self.name)
                if not ext or ext != item_ext:
                    os.makedirs(local_path, exist_ok=True)
                    local_path = os.path.join(local_path, self.name)

        # download
        filters = None
        items = self
        if self.type == 'dir':
            filters = self.datasets._bulid_folder_filter(folder_path=self.filename)
            items = None

        return self.items.download(items=items,
                                   local_path=local_path,
                                   file_types=file_types,
                                   save_locally=save_locally,
                                   to_array=to_array,
                                   annotation_options=annotation_options,
                                   overwrite=overwrite,
                                   to_items_folder=to_items_folder,
                                   annotation_filters=annotation_filters,
                                   thickness=thickness,
                                   alpha=alpha,
                                   with_text=with_text,
                                   export_version=export_version,
                                   filters=filters,
                                   dataset_lock=dataset_lock,
                                   lock_timeout_sec=lock_timeout_sec,
                                   export_summary=export_summary,
                                   raise_on_error=raise_on_error)

    def delete(self):
        """
        Delete item from platform

        :return: True
        :rtype: bool
        """
        return self.items.delete(item_id=self.id)

    def update(self, system_metadata=False):
        """
        Update items metadata

        :param bool system_metadata: bool - True, if you want to change metadata system
        :return: Item object
        :rtype: dtlpy.entities.item.Item
        """
        return self.items.update(item=self, system_metadata=system_metadata)

    def move(self, new_path):
        """
        Move item from one folder to another in Platform
        If the directory doesn't exist it will be created

        :param str new_path: new full path to move item to.
        :return: True if update successfully
        :rtype: bool
        """
        assert isinstance(new_path, str)
        if not new_path.startswith('/'):
            new_path = '/' + new_path
        if new_path.endswith('/'):
            self.filename = new_path + self.name
        else:
            try:
                self.items.get(filepath=new_path, is_dir=True)
                self.filename = new_path + '/' + self.name
            except exceptions.NotFound:
                self.filename = new_path

        return self.update(system_metadata=True)

    def clone(self, dst_dataset_id=None, remote_filepath=None, metadata=None, with_annotations=True,
              with_metadata=True, with_task_annotations_status=False, allow_many=False, wait=True):
        """
        Clone item

        :param str dst_dataset_id: destination dataset id
        :param str remote_filepath: complete filepath
        :param dict metadata: new metadata to add
        :param bool with_annotations: clone annotations
        :param bool with_metadata: clone metadata
        :param bool with_task_annotations_status: clone task annotations status
        :param bool allow_many: `bool` if True, using multiple clones in single dataset is allowed, (default=False)
        :param bool wait: wait for the command to finish
        :return: Item object
        :rtype: dtlpy.entities.item.Item

        **Example**:

        .. code-block:: python

            item.clone(item_id='item_id',
                    dst_dataset_id='dist_dataset_id',
                    with_metadata=True,
                    with_task_annotations_status=False,
                    with_annotations=False)
        """
        if remote_filepath is None:
            remote_filepath = self.filename
        if dst_dataset_id is None:
            dst_dataset_id = self.dataset_id
        return self.items.clone(item_id=self.id,
                                dst_dataset_id=dst_dataset_id,
                                remote_filepath=remote_filepath,
                                metadata=metadata,
                                with_annotations=with_annotations,
                                with_metadata=with_metadata,
                                with_task_annotations_status=with_task_annotations_status,
                                allow_many=allow_many,
                                wait=wait)

    def open_in_web(self):
        """
        Open the items in web platform

        :return:
        """
        self._client_api._open_in_web(url=self.platform_url)

    def _set_action(self, status: str, operation: str, assignment_id: str = None, task_id: str = None):
        """
        update item status

        :param status: str - string the describes the status
        :param operation: str -  'create' or 'delete'
        :param assignment_id: str - assignment id
        :param task_id: str - task id

        :return :True/False
        """
        if assignment_id:
            success = self.assignments.set_status(
                status=status,
                operation=operation,
                item_id=self.id,
                assignment_id=assignment_id
            )
        elif task_id:
            success = self.tasks.set_status(
                status=status,
                operation=operation,
                item_ids=[self.id],
                task_id=task_id
            )

        else:
            raise exceptions.PlatformException('400', 'Must provide task_id or assignment_id')

        return success

    def update_status(self, status: str, clear: bool = False, assignment_id: str = None, task_id: str = None):
        """
        update item status

        :param str status: "completed" ,"approved" ,"discard"
        :param bool clear: if true delete status
        :param str assignment_id: assignment id
        :param str task_id: task id

        :return :True/False
        :rtype: bool

        **Example**:

        .. code-block:: python

            item.update_status(status='complete',
                               task_id='task_id')

        """
        if not assignment_id and not task_id:
            system_metadata = self.metadata.get('system', dict())
            if 'refs' in system_metadata:
                refs = system_metadata['refs']
                if len(refs) <= 2:
                    for ref in refs:
                        if ref.get('type', '') == 'assignment':
                            assignment_id = ref['id']
                        if ref.get('type', '') == 'task':
                            task_id = ref['id']

        if assignment_id or task_id:
            if clear:
                self._set_action(status=status, operation='delete', assignment_id=assignment_id, task_id=task_id)
            else:
                self._set_action(status=status, operation='create', assignment_id=assignment_id, task_id=task_id)
        else:
            raise exceptions.PlatformException('400', 'must provide assignment_id or task_id')

    def status(self, assignment_id: str = None, task_id: str = None):
        """
        Get item status

        :param str assignment_id: assignment id
        :param str task_id: task id

        :return: status
        :rtype: str

        **Example**:

        .. code-block:: python

            status = item.status(task_id='task_id')
        """
        if not assignment_id and not task_id:
            raise exceptions.PlatformException('400', 'must provide assignment_id or task_id')
        status = None
        resource_id = assignment_id if assignment_id else task_id
        for ref in self.metadata.get('system', dict()).get('refs', []):
            if ref.get('id') == resource_id:
                status = ref.get('metadata', {}).get('status', None)
                break
        return status

    def set_description(self, text: str):
        """
        Update Item description

        :param str text: if None or "" description will be deleted

        :return
        """
        if text is None:
            text = ""
        if not isinstance(text, str):
            raise ValueError("Description must get string")
        self._description = text
        self._platform_dict = self.update()._platform_dict
        return self

    def assign_subset(self, subset: str):
        """
        Assign a single ML subset (train/validation/test) to this item.
        Sets the chosen subset to True and the others to None.
        Then calls item.update(system_metadata=True).

        :param str subset: 'train', 'validation', or 'test'
        """
        if subset not in ['train', 'validation', 'test']:
            raise ValueError("subset must be one of: 'train', 'validation', 'test'")

        if 'system' not in self.metadata:
            self.metadata['system'] = {}
        if 'tags' not in self.metadata['system']:
            self.metadata['system']['tags'] = {}

        self.metadata['system']['tags']['train'] = True if subset == 'train' else None
        self.metadata['system']['tags']['validation'] = True if subset == 'validation' else None
        self.metadata['system']['tags']['test'] = True if subset == 'test' else None

        return self.update(system_metadata=True)


    def remove_subset(self):
        """
        Remove any ML subset assignment from this item.
        Sets train, validation, and test to None.
        Then calls item.update(system_metadata=True).
        """
        if 'system' not in self.metadata:
            self.metadata['system'] = {}
        if 'tags' not in self.metadata['system']:
            self.metadata['system']['tags'] = {}

        self.metadata['system']['tags']['train'] = None
        self.metadata['system']['tags']['validation'] = None
        self.metadata['system']['tags']['test'] = None

        return self.update(system_metadata=True)


    def get_current_subset(self) -> str:
        """
        Get the current ML subset assignment of this item.
        Returns 'train', 'validation', 'test', or None if not assigned.

        :return: subset name or None
        :rtype: str or None
        """
        tags = self.metadata.get('system', {}).get('tags', {})
        for subset in ['train', 'validation', 'test']:
            if tags.get(subset) is True:
                return subset
        return None
    
    def assign_collection(self, collections: List[str]) -> bool:
        """
        Assign this item to one or more collections.

        :param collections: List of collection names to assign the item to.
        :return: True if the assignment was successful, otherwise False.
        """
        return self.collections.assign(dataset_id=self.dataset_id, collections=collections, item_id=self.id,)

    def unassign_collection(self, collections: List[str]) -> bool:
        """
        Unassign this item from one or more collections.

        :param collections: List of collection names to unassign the item from.
        :return: True if the unassignment was successful, otherwise False.
        """
        return self.collections.unassign(dataset_id=self.dataset_id, item_id=self.id, collections=collections)

    def list_collections(self) -> List[dict]:
        """
        List all collections associated with this item.

        :return: A list of dictionaries containing collection keys and their respective names.
                Each dictionary has the structure: {"key": <collection_key>, "name": <collection_name>}.
        """
        collections = self.metadata.get("system", {}).get("collections", {})
        if not isinstance(collections, dict):
            # Ensure collections is a dictionary
            return []
    
        # Retrieve collection names by their keys
        return [
            {"key": key, "name": self.collections.get_name_by_key(key)}
            for key in collections.keys()
        ]
    
    def task_scores(self, task_id: str, page_offset: int = None, page_size: int = None):
        """
        Get the scores of the item in a specific task.
        :param task_id: The ID of the task.
        :return: page of scores
        """
        return self.items.task_scores(item_id=self.id, task_id=task_id, page_offset=page_offset, page_size=page_size)


class ModalityTypeEnum(str, Enum):
    """
    State enum
    """
    OVERLAY = "overlay"
    REPLACE = "replace"
    PREVIEW = "preview"


class ModalityRefTypeEnum(str, Enum):
    """
    State enum
    """
    ID = "id"
    URL = "url"


class Modality:
    def __init__(self, _json=None, modality_type=None, ref=None, ref_type=ModalityRefTypeEnum.ID,
                 name=None, timestamp=None, mimetype=None):
        """
        :param _json: json represent of all modality params
        :param modality_type: ModalityTypeEnum.OVERLAY,ModalityTypeEnum.REPLACE
        :param ref: id or url of the item reference
        :param ref_type: ModalityRefTypeEnum.ID, ModalityRefTypeEnum.URL
        :param name:
        :param timestamp: ISOString, epoch of UTC
        :param mimetype: str - type of the file
        """
        if _json is None:
            _json = dict()
        self.type = _json.get('type', modality_type)
        self.ref_type = _json.get('refType', ref_type)
        self.ref = _json.get('ref', ref)
        self.name = _json.get('name', name)
        self.timestamp = _json.get('timestamp', timestamp)
        self.mimetype = _json.get('mimetype', mimetype)

    def to_json(self):
        _json = {"type": self.type,
                 "ref": self.ref,
                 "refType": self.ref_type}
        if self.name is not None:
            _json['name'] = self.name
        if self.timestamp is not None:
            _json['timestamp'] = self.timestamp
        if self.mimetype is not None:
            _json['mimetype'] = self.mimetype
        return _json


class Modalities:
    def __init__(self, item):
        assert isinstance(item, Item)
        self.item = item
        if 'system' not in self.item.metadata:
            self.item.metadata['system'] = dict()

    @property
    def modalities(self):
        mod = None
        if 'system' in self.item.metadata:
            mod = self.item.metadata['system'].get('modalities', None)
        return mod

    def create(self, name, ref,
               ref_type: ModalityRefTypeEnum = ModalityRefTypeEnum.ID,
               modality_type: ModalityTypeEnum = ModalityTypeEnum.OVERLAY,
               timestamp=None,
               mimetype=None,
               ):
        """
        create Modalities entity

        :param name: name
        :param ref: id or url of the item reference
        :param ref_type: ModalityRefTypeEnum.ID, ModalityRefTypeEnum.URL
        :param modality_type: ModalityTypeEnum.OVERLAY,ModalityTypeEnum.REPLACE
        :param timestamp: ISOString, epoch of UTC
        :param mimetype: str - type of the file
        """
        if self.modalities is None:
            self.item.metadata['system']['modalities'] = list()

        _json = {"type": modality_type,
                 "ref": ref,
                 "refType": ref_type}
        if name is not None:
            _json['name'] = name
        if timestamp is not None:
            _json['timestamp'] = timestamp
        if mimetype is not None:
            _json['mimetype'] = mimetype

        self.item.metadata['system']['modalities'].append(_json)

        return Modality(_json=_json)

    def delete(self, name):
        """
        :param name:
        """
        if self.modalities is not None:
            for modality in self.item.metadata['system']['modalities']:
                if name == modality['name']:
                    self.item.metadata['system']['modalities'].remove(modality)
                    return Modality(_json=modality)
        return None

    def list(self):
        modalities = list()
        if self.modalities is not None:
            modalities = list()
            for modality in self.item.metadata['system']['modalities']:
                modalities.append(Modality(_json=modality))
        return modalities



def main(first_project_name, second_project_name, first_dataset_name, second_dataset_name):
    """
    Copy folder from project/dataset/folder
    :return:
    """
    import dtlpy as dl

    # Get source project and dataset
    project = dl.projects.get(project_name=first_project_name)
    dataset_from = project.datasets.get(dataset_name=first_dataset_name)
    # filter to get all files of a specific folder
    filters = dl.Filters()
    filters.add(field='type', values='file')  # get only files
    filters.add(field='filename', values='/source_folder/**')  # get all items in folder (recursive)
    pages = dataset_from.items.list(filters=filters)

    # Get destination project and annotations
    project = dl.projects.get(project_name=second_project_name)
    dataset_to = project.datasets.get(dataset_name=second_dataset_name)

    # go over all projects and copy file from src to dst
    for page in pages:
        for item in page:
            # download item (without save to disk)
            buffer = item.download(save_locally=False)
            # give the items name to the buffer
            buffer.name = item.name
            # upload item
            new_item = dataset_to.items.upload(local_path=buffer,
                                               remote_path='/destination_folder')
            print(new_item.filename)



import requests
import logging
import base64
import enum
import json
import io
import os
from typing import List, Optional

from concurrent.futures import ThreadPoolExecutor
from .. import entities, repositories
from dtlpy.services.api_client import client as client_api

logger = logging.getLogger(name='dtlpy')


class PromptType(str, enum.Enum):
    TEXT = 'application/text'
    IMAGE = 'image/*'
    AUDIO = 'audio/*'
    VIDEO = 'video/*'
    METADATA = 'metadata'


class Prompt:
    def __init__(self, key, role='user'):
        """
        Create a single Prompt. Prompt can contain multiple mimetype elements, e.g. text sentence and an image.
        :param key: unique identifier of the prompt in the item
        """
        self.key = key
        self.elements = list()
        # to avoid broken stream of json files - DAT-75653
        client_api.default_headers['x-dl-sanitize'] = '0'
        self._items = repositories.Items(client_api=client_api)
        self.metadata = {'role': role}

    def add_element(self, value, mimetype='application/text'):
        """

        :param value: url or string of the input
        :param mimetype: mimetype of the input. options: `text`, `image/*`, `video/*`, `audio/*`
        :return:
        """
        allowed_prompt_types = [prompt_type for prompt_type in PromptType]
        if mimetype not in allowed_prompt_types:
            raise ValueError(f'Invalid mimetype: {mimetype}. Allowed values: {allowed_prompt_types}')
        if mimetype == PromptType.METADATA and isinstance(value, dict):
            self.metadata.update(value)
        else:
            self.elements.append({'mimetype': mimetype,
                                  'value': value})

    def to_json(self):
        """
        Convert Prompt entity to the item json

        :return:
        """
        elements_json = [
            {
                "mimetype": e['mimetype'],
                "value": e['value'],
            } for e in self.elements if not e['mimetype'] == PromptType.METADATA
        ]
        elements_json.append({
            "mimetype": PromptType.METADATA,
            "value": self.metadata
        })
        return {
            self.key: elements_json
        }

    def _convert_stream_to_binary(self, image_url: str):
        """
        Convert a stream to binary
        :param image_url: dataloop image stream url
        :return: binary object
        """
        image_buffer = None
        if '.' in image_url and 'dataloop.ai' not in image_url:
            # URL and not DL item stream
            try:
                response = requests.get(image_url, stream=True)
                response.raise_for_status()  # Raise an exception for bad status codes

                # Check for valid image content type
                if response.headers["Content-Type"].startswith("image/"):
                    # Read the image data in chunks to avoid loading large images in memory
                    image_buffer = b"".join(chunk for chunk in response.iter_content(1024))
            except requests.exceptions.RequestException as e:
                logger.error(f"Failed to download image from URL: {image_url}, error: {e}")

        elif '.' in image_url and 'stream' in image_url:
            # DL Stream URL
            item_id = image_url.split("/stream")[0].split("/items/")[-1]
            image_buffer = self._items.get(item_id=item_id).download(save_locally=False).getvalue()
        else:
            # DL item ID
            image_buffer = self._items.get(item_id=image_url).download(save_locally=False).getvalue()

        if image_buffer is not None:
            encoded_image = base64.b64encode(image_buffer).decode()
        else:
            logger.error(f'Invalid image url: {image_url}')
            return None

        return f'data:image/jpeg;base64,{encoded_image}'

    def messages(self):
        """
        return a list of messages in the prompt item,
        messages are returned following the openai SDK format https://platform.openai.com/docs/guides/vision
        """
        messages = []
        for element in self.elements:
            if element['mimetype'] == PromptType.TEXT:
                data = {
                    "type": "text",
                    "text": element['value']
                }
                messages.append(data)
            elif element['mimetype'] == PromptType.IMAGE:
                image_url = self._convert_stream_to_binary(element['value'])
                data = {
                    "type": "image_url",
                    "image_url": {
                        "url": image_url
                    }
                }
                messages.append(data)
            elif element['mimetype'] == PromptType.AUDIO:
                raise NotImplementedError('Audio prompt is not supported yet')
            elif element['mimetype'] == PromptType.VIDEO:
                data = {
                    "type": "video_url",
                    "video_url": {
                        "url": element['value']
                    }
                }
                messages.append(data)
            else:
                raise ValueError(f'Invalid mimetype: {element["mimetype"]}')
        return messages, self.key


class PromptItem:
    def __init__(self, name, item: entities.Item = None, role_mapping=None):
        if role_mapping is None:
            role_mapping = {'user': 'item',
                            'assistant': 'annotation'}
        if not isinstance(role_mapping, dict):
            raise ValueError(f'input role_mapping must be dict. type: {type(role_mapping)}')
        self.role_mapping = role_mapping
        # prompt item name
        self.name = name
        # list of user prompts in the prompt item
        self.prompts = list()
        self.assistant_prompts = list()
        # list of assistant (annotations) prompts in the prompt item
        # Dataloop Item
        self._item: entities.Item = item
        self._messages = []
        self._annotations: entities.AnnotationCollection = None
        if item is not None:
            if 'json' not in item.mimetype or item.system.get('shebang', dict()).get('dltype') != 'prompt':
                raise ValueError('Expecting a json item with system.shebang.dltype = prompt')
            self._items = item.items
            self.fetch()
        else:
            self._items = repositories.Items(client_api=client_api)

        # to avoid broken stream of json files - DAT-75653
        self._items._client_api.default_headers['x-dl-sanitize'] = '0'

    @classmethod
    def from_messages(cls, messages: list):
        ...

    @classmethod
    def from_item(cls, item: entities.Item):
        """
        Load a prompt item from the platform
        :param item : Item object
        :return: PromptItem object
        """
        if 'json' not in item.mimetype or item.system.get('shebang', dict()).get('dltype') != 'prompt':
            raise ValueError('Expecting a json item with system.shebang.dltype = prompt')
        return cls(name=item.name, item=item)

    @classmethod
    def from_local_file(cls, filepath):
        """
        Create a new prompt item from a file
        :param filepath: path to the file
        :return: PromptItem object
        """
        if os.path.exists(filepath) is False:
            raise FileNotFoundError(f'File does not exists: {filepath}')
        if 'json' not in os.path.splitext(filepath)[-1]:
            raise ValueError(f'Expected path to json item, got {os.path.splitext(filepath)[-1]}')
        prompt_item = cls(name=filepath)
        with open(filepath, 'r', encoding='utf-8') as f:
            data = json.load(f)
        prompt_item.prompts = prompt_item._load_item_prompts(data=data)
        return prompt_item

    @staticmethod
    def _load_item_prompts(data):
        prompts = list()
        for prompt_key, prompt_elements in data.get('prompts', dict()).items():
            content = list()
            for element in prompt_elements:
                content.append({'value': element.get('value', dict()),
                                'mimetype': element['mimetype']})
            prompt = Prompt(key=prompt_key, role="user")
            for element in content:
                prompt.add_element(value=element.get('value', ''),
                                   mimetype=element.get('mimetype', PromptType.TEXT))
            prompts.append(prompt)
        return prompts

    @staticmethod
    def _load_annotations_prompts(annotations: entities.AnnotationCollection):
        """
        Get all the annotations in the item for the assistant messages
        """
        # clearing the assistant prompts from previous annotations that might not belong
        assistant_prompts = list()
        for annotation in annotations:
            prompt_id = annotation.metadata.get('system', dict()).get('promptId', None)
            model_info = annotation.metadata.get('user', dict()).get('model', dict())
            annotation_id = annotation.id
            if annotation.type == 'ref_image':
                prompt = Prompt(key=prompt_id, role='assistant')
                prompt.add_element(value=annotation.annotation_definition.coordinates.get('ref'),
                                   mimetype=PromptType.IMAGE)
            elif annotation.type == 'text':
                prompt = Prompt(key=prompt_id, role='assistant')
                prompt.add_element(value=annotation.annotation_definition.coordinates,
                                   mimetype=PromptType.TEXT)
            else:
                raise ValueError(f"Unsupported annotation type: {annotation.type}")

            prompt.add_element(value={'id': annotation_id,
                                      'model_info': model_info},
                               mimetype=PromptType.METADATA)
            assistant_prompts.append(prompt)
        return assistant_prompts

    def to_json(self):
        """
        Convert the entity to a platform item.

        :return:
        """
        prompts_json = {
            "shebang": "dataloop",
            "metadata": {
                "dltype": 'prompt'
            },
            "prompts": {}
        }
        for prompt in self.prompts:
            for prompt_key, prompt_values in prompt.to_json().items():
                prompts_json["prompts"][prompt_key] = prompt_values
        return prompts_json

    def to_messages(self, model_name=None, include_assistant=True):
        all_prompts_messages = dict()
        for prompt in self.prompts:
            if prompt.key not in all_prompts_messages:
                all_prompts_messages[prompt.key] = list()
            prompt_messages, prompt_key = prompt.messages()
            messages = {
                'role': prompt.metadata.get('role', 'user'),
                'content': prompt_messages
            }
            all_prompts_messages[prompt.key].append(messages)
        if include_assistant is True:
            # reload to filer model annotations
            for prompt in self.assistant_prompts:
                prompt_model_name = prompt.metadata.get('model_info', dict()).get('name')
                if model_name is not None and prompt_model_name != model_name:
                    continue
                if prompt.key not in all_prompts_messages:
                    logger.warning(
                        f'Prompt key {prompt.key} is not found in the user prompts, skipping Assistant prompt')
                    continue
                prompt_messages, prompt_key = prompt.messages()
                assistant_messages = {
                    'role': 'assistant',
                    'content': prompt_messages
                }
                all_prompts_messages[prompt.key].append(assistant_messages)
        res = list()
        for prompts in all_prompts_messages.values():
            for prompt in prompts:
                res.append(prompt)
        self._messages = res
        return self._messages

    def to_bytes_io(self):
        # Used for item upload, do not delete
        byte_io = io.BytesIO()
        byte_io.name = self.name
        byte_io.write(json.dumps(self.to_json()).encode())
        byte_io.seek(0)
        return byte_io

    def fetch(self):
        if self._item is None:
            raise ValueError('Missing item, nothing to fetch..')
        self._item = self._items.get(item_id=self._item.id)
        self._annotations = self._item.annotations.list()
        self.prompts = self._load_item_prompts(data=json.load(self._item.download(save_locally=False)))
        self.assistant_prompts = self._load_annotations_prompts(self._annotations)

    def build_context(self, nearest_items, add_metadata=None) -> str:
        """
        Create a context stream from nearest items list.
        add_metadata is a list of location in the item.metadata to add to the context, for instance ['system.document.source']
        :param nearest_items: list of item ids
        :param add_metadata: list of metadata location to add metadata to context
        :return:
        """
        if add_metadata is None:
            add_metadata = list()

        def stream_single(w_id):
            context_item = self._items.get(item_id=w_id)
            buf = context_item.download(save_locally=False)
            text = buf.read().decode(encoding='utf-8')
            m = ""
            for path in add_metadata:
                parts = path.split('.')
                value = context_item.metadata
                part = ""
                for part in parts:
                    if isinstance(value, dict):
                        value = value.get(part)
                    else:
                        value = ""

                m += f"{part}:{value}\n"
            return text, m

        pool = ThreadPoolExecutor(max_workers=32)
        context = ""
        if len(nearest_items) > 0:
            # build context
            results = pool.map(stream_single, nearest_items)
            for res in results:
                context += f"\n<source>\n{res[1]}\n</source>\n<text>\n{res[0]}\n</text>"
        return context

    def add(self,
            message: dict,
            prompt_key: str = None,
            model_info: dict = None):
        """
        add a prompt to the prompt item
        prompt: a dictionary. keys are prompt message id, values are prompt messages
        responses: a list of annotations representing responses to the prompt

        :param message:
        :param prompt_key:
        :param model_info:
        :return:
        """
        role = message.get('role', 'user')
        content = message.get('content', list())

        if self.role_mapping.get(role, 'item') == 'item':
            if prompt_key is None:
                prompt_key = str(len(self.prompts) + 1)
            # for new prompt we need a new key
            prompt = Prompt(key=prompt_key, role=role)
            for element in content:
                prompt.add_element(value=element.get('value', ''),
                                   mimetype=element.get('mimetype', PromptType.TEXT))

            # create new prompt and add to prompts
            self.prompts.append(prompt)
            if self._item is not None:
                self._item._Item__update_item_binary(_json=self.to_json())
        else:
            if prompt_key is None:
                prompt_key = str(len(self.prompts))
            assistant_message = content[0]
            assistant_mimetype = assistant_message.get('mimetype', PromptType.TEXT)
            uploaded_annotation = None

            # find if prompt
            if model_info is None:
                # dont search for existing if there's no model information
                existing_prompt = None
            else:
                existing_prompts = list()
                for prompt in self.assistant_prompts:
                    prompt_id = prompt.key
                    model_name = prompt.metadata.get('model_info', dict()).get('name')
                    if prompt_id == prompt_key and model_name == model_info.get('name'):
                        # TODO how to handle multiple annotations
                        existing_prompts.append(prompt)
                if len(existing_prompts) > 1:
                    assert False, "shouldn't be here! more than 1 annotation for a single model"
                elif len(existing_prompts) == 1:
                    # found model annotation to upload
                    existing_prompt = existing_prompts[0]
                else:
                    # no annotation found
                    existing_prompt = None

            if existing_prompt is None:
                prompt = Prompt(key=prompt_key)
                if assistant_mimetype == PromptType.TEXT:
                    annotation_definition = entities.FreeText(text=assistant_message.get('value'))
                    prompt.add_element(value=annotation_definition.to_coordinates(None),
                                       mimetype=PromptType.TEXT)
                elif assistant_mimetype == PromptType.IMAGE:
                    annotation_definition = entities.RefImage(ref=assistant_message.get('value'))
                    prompt.add_element(value=annotation_definition.to_coordinates(None).get('ref'),
                                       mimetype=PromptType.IMAGE)
                else:
                    raise NotImplementedError('Only images of mimetype image and text are supported')
                metadata = {'system': {'promptId': prompt_key},
                            'user': {'model': model_info}}
                prompt.add_element(mimetype=PromptType.METADATA,
                                   value={"model_info": model_info})

                existing_annotation = entities.Annotation.new(item=self._item,
                                                              metadata=metadata,
                                                              annotation_definition=annotation_definition)
                uploaded_annotation = existing_annotation.upload()
                prompt.add_element(mimetype=PromptType.METADATA,
                                   value={"id": uploaded_annotation.id})
                existing_prompt = prompt
                self.assistant_prompts.append(prompt)

            existing_prompt_element = [element for element in existing_prompt.elements if
                                       element['mimetype'] != PromptType.METADATA][-1]
            existing_prompt_element['value'] = assistant_message.get('value')
            if uploaded_annotation is None:
                # Creating annotation with old dict to match platform dict
                annotation_definition = entities.FreeText(text='')
                metadata = {'system': {'promptId': prompt_key},
                            'user': {'model': existing_prompt.metadata.get('model_info')}}
                annotation = entities.Annotation.new(item=self._item,
                                                     metadata=metadata,
                                                     annotation_definition=annotation_definition
                                                     )
                annotation.id = existing_prompt.metadata['id']
                # set the platform dict to match the old annotation for the dict difference check, otherwise it won't
                # update
                annotation._platform_dict = annotation.to_json()
                # update the annotation with the new text
                annotation.annotation_definition.text = existing_prompt_element['value']
                self._item.annotations.update(annotation)

    def update(self):
        """
        Update the prompt item in the platform. 
        """
        if self._item is not None:
            self._item._Item__update_item_binary(_json=self.to_json())
            self._item = self._item.update()
        else:
            raise ValueError('Cannot update PromptItem without an item.')

    # Properties
    @property
    def item(self) -> Optional['entities.Item']:
        """
        Get the underlying Item object.

        :return: The Item object associated with this PromptItem, or None.
        :rtype: Optional[dtlpy.entities.Item]
        """
        return self._item

    @item.setter
    def item(self, item: Optional['entities.Item']):
        """
        Set the underlying Item object.

        :param item: The Item object to associate with this PromptItem, or None.
        :type item: Optional[dtlpy.entities.Item]
        """
        if item is not None and not isinstance(item, entities.Item):
            raise ValueError(f"Expected dtlpy.entities.Item or None, got {type(item)}")
        self._item = item


    @property
    def metadata(self) -> dict:
        """
        Get the metadata from the underlying Item object.

        :return: Metadata dictionary from the item, or empty dict if no item exists.
        :rtype: dict
        """
        if self._item is not None:
            return self._item.metadata
        else:
            raise ValueError('No item found, cannot get metadata, to set item use prompt_item.item = item')


from dtlpy import miscellaneous

from ..services.api_client import ApiClient
from .. import exceptions, entities
import logging

logger = logging.getLogger(name='dtlpy')

class ServiceDrivers:

    def __init__(self, client_api: ApiClient):
        self._client_api = client_api
        self._base_url = '/serviceDrivers'

    def create(
            self,
            name: str,
            compute_id: str,
            context: entities.ComputeContext,
            namespace: str = None
    ):
        """
        Create a new service driver

        :param name: Service driver name
        :param compute_id: Compute ID
        :param context: Compute context
        :param namespace: Namespace
        :return: Service driver

        """

        payload = {
            'name': name,
            'computeId': compute_id,
            'context': context.to_json()
        }
        if namespace is not None:
            payload['namespace'] = namespace

        # request
        success, response = self._client_api.gen_request(
            req_type='post',
            path=self._base_url,
            json_req=payload
        )

        if not success:
            raise exceptions.PlatformException(response)

        service_driver = entities.ServiceDriver.from_json(
            _json=response.json(),
            client_api=self._client_api
        )

        return service_driver

    def get(self, service_driver_id: str):
        """
        Get a service driver

        :param service_driver_id: Service driver ID
        :return: Service driver
        """

        # request
        success, response = self._client_api.gen_request(
            req_type='get',
            path=self._base_url + '/{}'.format(service_driver_id)
        )

        if not success:
            raise exceptions.PlatformException(response)

        service_driver = entities.ServiceDriver.from_json(
            _json=response.json(),
            client_api=self._client_api
        )

        return service_driver

    def delete(self, service_driver_id: str):
        """
        Delete a service driver

        :param service_driver_id: Service driver ID
        """

        # request
        success, response = self._client_api.gen_request(
            req_type='delete',
            path=self._base_url + '/{}'.format(service_driver_id)
        )

        if not success:
            raise exceptions.PlatformException(response)

        return True

    def set_default(self, service_driver_id: str, org_id: str, update_existing_services=False):
        """
        Set a service driver as default

        :param service_driver_id: Compute name
        :param org_id: Organization ID
        :param update_existing_services: Update existing services

        :return: Service driver
        """

        # request
        success, response = self._client_api.gen_request(
            req_type='post',
            path=self._base_url + '/default',
            json_req={
                'organizationId': org_id,
                'updateExistingServices': update_existing_services,
                'driverName': service_driver_id
            }
        )

        if not success:
            raise exceptions.PlatformException(response)

        service_driver = entities.ServiceDriver.from_json(
            _json=response.json(),
            client_api=self._client_api
        )

        return service_driver

    def update(self, service_driver: entities.ServiceDriver):
        """
        Update a service driver

        :param dtlpy.entities.service_driver.ServiceDriver service_driver: updated service driver object
        :return: Service driver entity
        :rtype: dtlpy.entities.service_driver.ServiceDriver
        """

        # request
        success, response = self._client_api.gen_request(
            req_type='patch',
            path=self._base_url + '/{}'.format(service_driver.id),
            json_req=service_driver.to_json()
        )

        if not success:
            raise exceptions.PlatformException(response)

        service_driver = entities.ServiceDriver.from_json(
            _json=response.json(),
            client_api=self._client_api
        )

        return service_driver

    def _list(self, filters: entities.Filters):
        url = self._base_url + '/query'
        success, response = self._client_api.gen_request(req_type='POST',
                                                         path=url,
                                                         json_req=filters.prepare())
        if not success:
            raise exceptions.PlatformException(response)

        return response.json()

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.ServiceDriver]:
        pool = self._client_api.thread_pools(pool_name='entity.create')
        jobs = [None for _ in range(len(response_items))]
        for i_item, item in enumerate(response_items):
            jobs[i_item] = pool.submit(entities.ServiceDriver._protected_from_json,
                                       **{'client_api': self._client_api,
                                          '_json': item})
        results = [j.result() for j in jobs]
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        items = miscellaneous.List([r[1] for r in results if r[0] is True])
        return items

    def list(self, filters: entities.Filters = None) -> entities.PagedEntities:
        """
        List all services drivers

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :return: Paged entity
        :rtype: dtlpy.entities.paged_entities.PagedEntities

        **Example**:

        .. code-block:: python

            services = dl.service_drivers.list()
        """
        # default filters
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.SERVICE_DRIVER)

        if filters.resource != entities.FiltersResource.SERVICE_DRIVER:
            raise exceptions.PlatformException(
                error='400',
                message='Filters resource must to be FiltersResource.SERVICE_DRIVER. Got: {!r}'.format(
                    filters.resource))

        if not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException('400', 'Unknown filters type')

        paged = entities.PagedEntities(items_repository=self,
                                       filters=filters,
                                       page_offset=filters.page,
                                       page_size=filters.page_size,
                                       client_api=self._client_api)
        paged.get_page()
        return paged



def main():
    """

    :return:
    """
    import dtlpy as dl

    # Get project and dataset
    project = dl.projects.get(project_name='Curling')
    dataset = project.datasets.get(dataset_name='Practice')

    # upload specific files
    dataset.items.upload(local_path=['/home/project/images/John Morris.jpg',
                                     '/home/project/images/John Benton.jpg',
                                     '/home/project/images/Liu Jinli.jpg'],
                         remote_path='/first_batch')

    # upload all files in a folder
    dataset.items.upload(local_path='/home/project/images',
                         remote_path='/first_batch')



import threading
import asyncio
import logging
import io

logger = logging.getLogger(name='dtlpy')


class AsyncThreadEventLoop(threading.Thread):
    """Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition."""

    def __init__(self, loop, n, *args, **kwargs):
        super(AsyncThreadEventLoop, self).__init__(*args, **kwargs)
        self.loop = loop
        self.n = n
        self._count = 0
        self._semaphores = dict()

    def count_up(self):
        self._count += 1

    def count_down(self):
        self._count -= 1

    def run(self):
        def exception_handler(loop, context):
            logger.debug(
                "[Asyc] EventLoop: caught the following exception: {}".format(context['message']))

        self.loop.set_exception_handler(exception_handler)
        asyncio.set_event_loop(self.loop)
        self.loop.run_forever()
        logger.debug('Ended event loop with bounded semaphore to {}'.format(self.n))

    def semaphore(self, name, n=None):
        if n is None:
            n = self.n
        else:
            n = min(n, self.n)
        if name not in self._semaphores:
            self._semaphores[name] = asyncio.BoundedSemaphore(n)
        return self._semaphores[name]

    def stop(self):
        self.loop.call_soon_threadsafe(self.loop.stop)  # here


class AsyncResponse:
    def __init__(self, text, _json, async_resp):
        self.text = text
        self._json = _json
        self.async_resp = async_resp

    def json(self):
        return self._json

    @property
    def status_code(self):
        return self.async_resp.status

    @property
    def reason(self):
        return self.async_resp.reason

    @property
    def ok(self):
        return self.async_resp.status < 400

    @property
    def request(self):
        return self.async_resp.request_info

    @property
    def headers(self):
        return self.async_resp.headers


class DummyErrorResponse:
    def __init__(self, error, trace):
        self.status = 3001  # SDK Error
        self.reason = trace
        self.error = error
        self.request_info = {}
        self.headers = {}


class AsyncResponseError(AsyncResponse):
    def __init__(self, error, trace):
        async_resp = DummyErrorResponse(error=error, trace=trace)
        _json = {'error': error}
        text = error
        super().__init__(async_resp=async_resp,
                         _json=_json,
                         text=text)


class AsyncUploadStream(io.IOBase):
    def __init__(self, buffer, callback=None, name='', chunk_timeout=10, max_retries=3):
        self.buffer = buffer
        self.buffer.seek(0)
        self.callback = callback
        self._name = name
        self.chunk_timeout = chunk_timeout
        self.max_retries = max_retries

    @property
    def name(self):
        return self._name

    async def async_read(self, size):
        retries = 0
        while retries < self.max_retries:
            try:
                import sys
                if sys.version_info < (3, 9):
                    loop = asyncio.get_event_loop()
                    data = await asyncio.wait_for(loop.run_in_executor(None, self.buffer.read, size),
                                                  timeout=self.chunk_timeout)
                else:
                    data = await asyncio.wait_for(asyncio.to_thread(self.buffer.read, size), timeout=self.chunk_timeout)
                if self.callback is not None:
                    self.callback(size)
                return data
            except asyncio.TimeoutError:
                retries += 1
                logger.warning(
                    f"Chunk read timed out after {self.chunk_timeout} seconds. Retrying {retries}/{self.max_retries}...")

        raise Exception(f"Chunk read failed after {self.max_retries} retries due to timeouts")

    def read(self, size):
        return asyncio.run(self.async_read(size))



from typing import Optional, Union
import pandas as pd
import numpy as np

NUMBER_TYPES = (int, float, complex)


class FigOptions:
    def __init__(self, rows_per_page: int = None, comments: dict = None, x_title: str = None, y_title: str = None,
                 colors: list = None, direction: str = None):
        self.rows_per_page = rows_per_page
        self.comments = comments
        self.x_title = x_title
        self.y_title = y_title
        self.colors = colors
        self.direction = direction

        self._validate()

    def _validate(self):
        # Mapping attribute names to their expected types
        expected_types = {
            'rows_per_page': int,
            'comments': dict,
            'x_title': str,
            'y_title': str,
            'colors': list,
            'direction': str
        }

        for attr, expected_type in expected_types.items():
            value = getattr(self, attr)
            if value is not None and not isinstance(value, expected_type):
                raise ValueError(
                    f"{attr} must be of type {expected_type.__name__}")

    def to_dict(self):
        return {
            "comments": self.comments,
            "rowsPerPage": self.rows_per_page,
            "xTitle": self.x_title,
            "yTitle": self.y_title,
            "colors": self.colors,
            "direction": self.direction,
        }


class Base:
    def __init__(self,
                 title: str,
                 type: str,
                 labels: list,
                 data: Union[list, np.ndarray],
                 title_href: Optional[str] = None,
                 plot_id: Optional[str] = None,
                 options: Optional[FigOptions] = None):
        self.title = title
        self.type = type
        self.title_href = title_href
        self.plot_id = plot_id
        self.labels = labels
        self.data = data
        self.options = options if options is not None and options != {} else FigOptions()

        self._validate()

    def _validate(self):
        if not isinstance(self.title, str):
            raise ValueError("title must be a string")

        if self.title_href is not None and not isinstance(self.title_href, str):
            raise ValueError("title_href must be a string or None")

        if self.plot_id is not None and not isinstance(self.plot_id, str):
            raise ValueError("plot_id must be a string or None")

        if not isinstance(self.labels, list):
            raise ValueError("labels must be a list")

        if not isinstance(self.data, (list, np.ndarray)):
            raise ValueError("data must be a list or a numpy ndarray")

        if not isinstance(self.options, FigOptions):
            raise ValueError("options must be an instance of FigOptions")

        if not isinstance(self.type, str):
            raise ValueError("type must be a string")

    def to_dict(self):
        options = self.options.to_dict()
        if isinstance(self, ConfusionMatrix):
            options["confusion"] = True

        data = self.data
        if isinstance(data, np.ndarray):
            data = data.tolist()

        return {
            "type": self.type,
            "title": self.title,
            "href": self.title_href,
            "labels": self.labels,
            "id": self.plot_id,
            "data": data,
            "options": options
        }


class Table(Base):
    def __init__(self,
                 title: str,
                 labels: list,
                 data: Union[list, np.ndarray],
                 title_href: Optional[str] = None,
                 plot_id: Optional[str] = None,
                 options: Optional[FigOptions] = None):
        super().__init__(title=title,
                         type="table",
                         labels=labels,
                         data=data,
                         title_href=title_href,
                         plot_id=plot_id,
                         options=options)
        self._validate_table_data()

    def _validate_table_data(self):
        data, labels = self.data, self.labels
        n_cols = len(labels)

        if not isinstance(data, list):
            raise ValueError(
                'arg "data" must be a list of lists. got a {!r}'.format(type(data)))

        for i_d, d in enumerate(data):
            if not isinstance(d, list):
                raise ValueError(
                    '"data" must be a list of lists. got type {!r} at index {}'.format(type(d), i_d))
            if len(d) != n_cols:
                raise ValueError(
                    '"data" rows must be the same size as "labels": {}. got size {} at index {}'.format(n_cols, len(d),
                                                                                                        i_d))

    @classmethod
    def from_df(cls, df: pd.DataFrame, **kwargs):
        labels = df.columns.to_list()
        data = df.values.tolist()
        return cls(labels=labels, data=data, **kwargs)


class Line(Base):
    def __init__(self,
                 title: str,
                 labels: list,
                 data: Union[list, np.ndarray],
                 title_href: Optional[str] = None,
                 plot_id: Optional[str] = None,
                 options: Optional[FigOptions] = None):
        super().__init__(title=title,
                         type="line",
                         labels=labels,
                         data=data,
                         title_href=title_href,
                         plot_id=plot_id,
                         options=options)


class Scatter(Base):
    def __init__(self,
                 title: str,
                 data: Union[list, np.ndarray],
                 labels: list = None,
                 title_href: Optional[str] = None,
                 plot_id: Optional[str] = None,
                 options: Optional[FigOptions] = None):
        if labels is None:
            labels = list()
        super().__init__(title=title,
                         type='scatter',
                         labels=labels,
                         data=data,
                         title_href=title_href,
                         plot_id=plot_id,
                         options=options)


class Bar(Base):
    def __init__(self,
                 title: str,
                 labels: list,
                 data: Union[list, np.ndarray],
                 title_href: Optional[str] = None,
                 plot_id: Optional[str] = None,
                 options: Optional[FigOptions] = None):
        super().__init__(title=title,
                         type="bar",
                         labels=labels,
                         data=data,
                         title_href=title_href,
                         plot_id=plot_id,
                         options=options)


class Doughnut(Base):
    def __init__(self,
                 title: str,
                 labels: list,
                 data: Union[list, np.ndarray],
                 title_href: Optional[str] = None,
                 plot_id: Optional[str] = None,
                 options: Optional[FigOptions] = None):
        super().__init__(title=title,
                         type="doughnut",
                         labels=labels,
                         data=data,
                         title_href=title_href,
                         plot_id=plot_id,
                         options=options)


class Pie(Base):
    def __init__(self,
                 title: str,
                 labels: list,
                 data: Union[list, np.ndarray],
                 title_href: Optional[str] = None,
                 plot_id: Optional[str] = None,
                 options: Optional[FigOptions] = None):
        super().__init__(title=title,
                         type='pie',
                         labels=labels,
                         data=data,
                         title_href=title_href,
                         plot_id=plot_id,
                         options=options)
        self._validate_pie_data()

    def _validate_pie_data(self):
        data, labels = self.data, self.labels
        if not isinstance(data, list):
            raise ValueError(
                'arg "data" must be a list. got a {!r}'.format(type(data)))

        if len(data) != len(labels):
            raise ValueError('"data" must be the same size as "labels". got size {} and {}'.format(len(data),
                                                                                                   len(labels)))
        for i_d, d in enumerate(data):
            if not isinstance(d, NUMBER_TYPES):
                raise ValueError(
                    'all "data" fields must be number. got type {!r} at index {}'.format(type(d), i_d))


class Hbar(Base):
    def __init__(self,
                 title: str,
                 labels: list,
                 data: Union[list, np.ndarray],
                 title_href: Optional[str] = None,
                 plot_id: Optional[str] = None,
                 options: Optional[FigOptions] = None):
        super().__init__(title=title,
                         type="hbar",
                         labels=labels,
                         data=data,
                         title_href=title_href,
                         plot_id=plot_id,
                         options=options)


class ConfusionMatrix(Base):
    def __init__(self,
                 title: str,
                 labels: list,
                 data: Union[list, np.ndarray],
                 title_href: Optional[str] = None,
                 plot_id: Optional[str] = None,
                 options: Optional[FigOptions] = None,
                 href_map: Optional[list] = None,
                 color_map: Optional[np.ndarray] = None):
        self.xlabel = 'Predicted'
        self.ylabel = 'Actual'
        self.color_map = color_map
        self.href_map = href_map
        super().__init__(title=title,
                         type="table",
                         labels=labels,
                         data=data,
                         title_href=title_href,
                         plot_id=plot_id,
                         options=options)
        self._validate_confusion_matrix_data()

    def _validate_confusion_matrix_data(self):
        if self.color_map is not None and not isinstance(self.color_map, np.ndarray):
            raise ValueError('arg "color_map" must be a numpy ndarray. got a {!r}'.format(
                type(self.color_map)))
        if self.href_map is not None and not isinstance(self.href_map, list):
            raise ValueError(
                'arg "href_map" must be a list. got a {!r}'.format(type(self.href_map)))

    @staticmethod
    def _rbgtohex(r, g, b):
        return "#{0:02x}{1:02x}{2:02x}".format(int(255 * r), int(255 * g), int(255 * b))

    def to_dict(self):
        options = self.options.to_dict()
        options["confusion"] = True
        data = self.data.copy()
        if isinstance(data, np.ndarray):
            data = data.tolist()
        labels = self.labels.copy()
        for i_row, row in enumerate(data):
            for i_col, val in enumerate(row):
                color = self._rbgtohex(
                    *self.color_map[i_row, i_col, :3]) if self.color_map is not None else None
                href = self.href_map[i_row][i_col] if self.href_map is not None else None
                data[i_row][i_col] = {
                    'text': '{:.2f}'.format(val),
                    'href': href,
                    'color': color}
            row.insert(0, labels[i_row])
        labels.insert(0, f'{self.ylabel}//{self.xlabel}')
        return {
            "type": self.type,
            "title": self.title,
            "href": self.title_href,
            "labels": labels,
            "data": data,
            "options": options
        }
    # def from_df(self, df: pd.DataFrame, **kwargs):
    #     data = []
    #     for i_label, g_label in enumerate(labels):
    #         line = list()
    #         line.append(g_label)
    #         for r_label in labels:
    #             if g_label == r_label:
    #                 p = high_prob()
    #             else:
    #                 p = low_prob()
    #             line.append({'text': str(p),
    #                          'href': "https://rc-con.dataloop.ai/projects/2cb9ae90-b6e8-4d15-9016-17bacc9b7bdf/datasets/607ed8107370454e4dd3b4c7/items?view=icons&dqlFilter=%7B%22filter%22%3A+%7B%22%24and%22%3A+%5B%7B%22hidden%22%3A+false%7D%2C+%7B%22type%22%3A+%22file%22%7D%2C+%7B%22dir%22%3A+%22booking%22%7D%5D%7D%2C+%22page%22%3A+0%2C+%22pageSize%22%3A+1000%2C+%22resource%22%3A+%22items%22%2C+%22join%22%3A+%7B%22on%22%3A+%7B%22resource%22%3A+%22annotations%22%2C+%22local%22%3A+%22itemId%22%2C+%22forigen%22%3A+%22id%22%7D%2C+%22filter%22%3A+%7B%22%24and%22%3A+%5B%7B%22label%22%3A+%22immigration%22%7D%5D%7D%7D%7D",
    #                          'color': rbgtohex(*colors(p)[:3])})
    #         data.append(line)



from . import BaseAnnotationDefinition


class RefImage(BaseAnnotationDefinition):
    """
    Create an image annotation. Reference the url or item id in this annotation type
    """
    type = "ref_image"

    def __init__(self, ref, ref_type=None, mimetype=None, label='ref-image', attributes=None, description=None):

        """
        Create an image annotation. Used for generative model and any other algorithm where and image is the output

        For type 'id', need to upload the image as item in the platform and reference the item id in the annotation.
        For type 'url', mimetype must be provided to load the ref correctly in the platform

        :param str ref: the reference to the image annotation, represented by an ‘itemId’ or ‘url’
        :param str ref_type: one of ‘id’ | ‘url’
        :param str mimetype: optional. in case the refType is URL, e.g. image/jpeg, video/mpeg
        :param label: annotation label
        :param attributes: annotation attributes
        :param description:

        :return:
        """
        super().__init__(description=description, attributes=attributes)
        if ref_type is None:
            if ref.startswith('http'):
                ref_type = 'url'
            else:
                ref_type = 'id'
        self.ref = ref
        self.ref_type = ref_type
        self.mimetype = mimetype
        self.label = label

    @property
    def x(self):
        return 0

    @property
    def y(self):
        return 0

    @property
    def geo(self):
        return list()

    def show(self, image, thickness, with_text, height, width, annotation_format, color, alpha=1):
        """
        Show annotation as ndarray
        :param image: empty or image to draw on
        :param thickness:
        :param with_text: not required
        :param height: item height
        :param width: item width
        :param annotation_format: options: list(dl.ViewAnnotationOptions)
        :param color: color
        :param alpha: opacity value [0 1], default 1
        :return: ndarray
        """
        # TODO over or show the image annotations
        return self.add_text_to_image(image=image, annotation=self)

    def to_coordinates(self, color):
        coordinates = {
            "ref": self.ref,
            "refType": self.ref_type,
            "mimetype": self.mimetype,
        }
        return coordinates

    @classmethod
    def from_json(cls, _json):
        coordinates = _json["coordinates"]
        ref = coordinates.get('ref')
        ref_type = coordinates.get('refType')
        mimetype = coordinates.get('mimetype')
        return cls(
            ref=ref,
            ref_type=ref_type,
            mimetype=mimetype,
            label=_json["label"],
            attributes=_json.get("attributes", None),
        )



from .. import entities
from .. import PlatformException
from enum import Enum
import os
import mimetypes


class LinkTypeEnum(str, Enum):
    """
    State enum
    """
    ID = "id"
    URL = "url"


class Link:

    # noinspection PyShadowingBuiltins
    def __init__(self, name, type: LinkTypeEnum, ref, mimetype=None, dataset_id=None):
        self.type = type
        self.ref = ref
        self.name = name
        self.dataset_id = dataset_id
        self.mimetype = mimetype


class UrlLink(Link):

    def __init__(self, ref, mimetype=None, name=None):
        if name is None:
            name = os.path.split(ref)[-1].replace('/', "").replace("\\", '').replace('.', '')
            name = name.split('?')[0]
        # noinspection PyShadowingBuiltins
        type = LinkTypeEnum.URL
        if not mimetype:
            mimetype = mimetypes.guess_type(ref.split('?')[0])[0] or 'image'
        super().__init__(name=name, type=type, mimetype=mimetype, ref=ref)

    @staticmethod
    def from_list(url_list):
        url_links = list()
        for url in url_list:
            url_links.append(UrlLink(ref=url))
        return url_links


class ItemLink(Link):

    def __init__(self, ref=None, name=None, dataset_id=None, item=None):
        if ref is None and item is None:
            raise PlatformException('400', 'Must provide either ref or item_id')

        if item is not None and not isinstance(item, entities.Item):
            raise PlatformException('400', 'Param item must be of type Item')

        if ref is None:
            ref = item.id

        if name is None:
            if item is not None:
                name = item.name
            else:
                name = ref

        if dataset_id is None and item is not None and isinstance(item, entities.Item):
            dataset_id = item.dataset_id

        # noinspection PyShadowingBuiltins
        type = LinkTypeEnum.ID
        super().__init__(name=name, type=type, ref=ref, dataset_id=dataset_id)

    @staticmethod
    def from_list(items):
        if not isinstance(items, list):
            items = [items]

        item_links = list()
        for item in items:
            if isinstance(item, str):
                item_links.append(ItemLink(ref=item))
            elif isinstance(item, entities.Item):
                item_links.append(ItemLink(item=item))
            else:
                raise PlatformException('400', 'Unknown item type. items should be a list of item entities/item ids')
        return item_links



import functools
import logging
import enum

from .. import miscellaneous

logger = logging.getLogger('dtlpy')


class EntityScopeLevel(str, enum.Enum):
    PRIVATE = 'private',
    PROJECT = 'project',
    ORG = 'org',
    PUBLIC = 'public'


class BaseEntity(object):
    is_fetched = True

    def print(self, to_return=False, columns=None):
        """
        :param to_return:
        :param columns:
        """
        return miscellaneous.List([self]).print(to_return=to_return, columns=columns)

    def to_df(self, show_all=False, columns=None):
        """
        :param show_all:
        :param columns:
        """
        return miscellaneous.List([self]).to_df(show_all=show_all, columns=columns)


# for auto-complete
from typing import TYPE_CHECKING

RUNTIME = not TYPE_CHECKING
if RUNTIME:
    def base_model(model):
        return model
else:
    from dataclasses import dataclass as base_model


class DlList(list):
    def __init__(self, _list=None, _dict=None, _type=None):
        if _list is None:
            _list = list()
        if _dict is None:
            _dict = list()
        self._dict = _dict
        self._type = _type
        super(DlList, self).__init__(_list)

    def append(self, item) -> None:
        if self._type is not None:
            assert isinstance(item, self._type), f'Cannot append type: {type(item)}. Must be type {self._type}'
        self._dict.append(item.to_json())
        super(DlList, self).append(item)  # append the item to itself (the list)


@base_model
class DlEntity(object):
    is_fetched = True

    def __init__(self, _dict=None, **kwargs):
        # using dict by reference and not creating a new one each time
        if _dict is None:
            _dict = dict()
        self._dict = _dict
        # this will set all the inputs to right location (using the dl.Property definitions)
        for key, value in kwargs.items():
            if key == "_dict":
                self._dict.update(value.copy())
            else:
                self.__setattr__(key, value)
        self._set_defaults()

    def _set_defaults(self):
        # getting all attribute (that are DlProperty) to set defaults to _dict
        # https://stackoverflow.com/questions/21962769/is-it-possible-to-test-if-object-property-uses-a-descriptor
        for att, a_type in type(self).__mro__[0].__dict__.items():
            if isinstance(a_type, DlProperty):
                _ = self.__getattribute__(att)

    def print(self, to_return=False, columns=None):
        """
        :param to_return:
        :param columns:
        """
        return miscellaneous.List([self]).print(to_return=to_return, columns=columns)

    def to_df(self, show_all=False, columns=None):
        """
        :param show_all:
        :param columns:
        """
        return miscellaneous.List([self]).to_df(show_all=show_all, columns=columns)


class DlProperty:
    def __init__(self, location=None, default='NODEFAULT', _type=None, _kls=None):
        self.location = location
        self._default = default
        self._type = _type
        self._validator = None
        self._kls = _kls

    ##############
    # decorators #
    ##############
    def default(self, method=None):
        @functools.wraps(method)
        def _default(self):
            return method(self)

        self._default = _default
        return self._default

    def validator(self, method):
        @functools.wraps(method)
        def _validator(self, value):
            return method(self, value)

        self._validator = _validator
        return self._validator

    ############
    # privates #
    ############
    @staticmethod
    def _get_class(kls):
        import importlib
        module = importlib.import_module(f'.entities', package='dtlpy')
        kls = getattr(module, kls)
        return kls

    def _get_default(self, instance):
        if callable(self._default):
            default = self._default(self=instance)
        else:
            default = self._default
        return default

    def _to_dict(self, inst):
        if self._kls is not None:
            if isinstance(inst, list) and all(hasattr(v, 'to_json') for v in inst):
                _dict = [v.to_json() for v in inst]
            elif hasattr(inst, 'to_json'):
                _dict = inst.to_json()
            else:
                _dict = inst
        else:
            _dict = inst
        return _dict

    def _to_instance(self, _dict):
        if self._kls is not None:
            kls = self._get_class(kls=self._kls)
            if isinstance(_dict, list) and all(isinstance(v, dict) for v in _dict):
                value = DlList(_list=[kls.from_json(v) for v in _dict],
                               _type=kls,
                               _dict=_dict)
            elif isinstance(_dict, dict):
                value = kls(_dict=_dict, **_dict)
            else:
                value = _dict
        else:
            value = _dict
        return value

    def __get__(self, instance, owner):
        _dict = instance._dict
        for key in self.location[:-1]:
            _dict = _dict.get(key, dict())

        # get default if set - can also be callable
        try:
            value = _dict[self.location[-1]]
        except KeyError:
            # get the default only of value doesnt exists
            value = self._get_default(instance=instance)
            if value == 'NODEFAULT':
                # if NODEFAULT - set value to None and DONT save in the _dict
                value = None
            else:
                # if default value is set - add the value into the _dict
                _dict[self.location[-1]] = value

        # instantiate dictionary to the type
        value = self._to_instance(_dict=value)
        return value

    def __set__(self, instance, value):

        # validate - if validator is set
        if self._validator is not None:
            # instantiate the value before validation
            inst = self._to_instance(_dict=value)
            self._validator(self=instance, value=inst)

        # convert to the value's kls
        value = self._to_dict(inst=value)

        # set value in the dictionary
        _dict = instance._dict
        for key in self.location[:-1]:
            _tmp_dict = _dict.get(key, None)
            # create the dict inside to be able to point
            if _tmp_dict is None:
                _dict[key] = dict()
            _dict = _dict.get(key, None)
        _dict[self.location[-1]] = value
